import sys
import numpy,math
from scipy import interpolate
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation, MPDecorator
from schainpy.model.data.jrodata import Voltage,hildebrand_sekhon
from schainpy.utils import log
from time import time



class VoltageProc(ProcessingUnit):

    def __init__(self):

        ProcessingUnit.__init__(self)

        self.dataOut = Voltage()
        self.flip = 1
        self.setupReq = False

    def run(self):

        if self.dataIn.type == 'AMISR':
            self.__updateObjFromAmisrInput()

        if self.dataIn.type == 'Voltage':
            self.dataOut.copy(self.dataIn)

    def __updateObjFromAmisrInput(self):

        self.dataOut.timeZone = self.dataIn.timeZone
        self.dataOut.dstFlag = self.dataIn.dstFlag
        self.dataOut.errorCount = self.dataIn.errorCount
        self.dataOut.useLocalTime = self.dataIn.useLocalTime

        self.dataOut.flagNoData = self.dataIn.flagNoData
        self.dataOut.data = self.dataIn.data
        self.dataOut.utctime = self.dataIn.utctime
        self.dataOut.channelList = self.dataIn.channelList
        #self.dataOut.timeInterval = self.dataIn.timeInterval
        self.dataOut.heightList = self.dataIn.heightList
        self.dataOut.nProfiles = self.dataIn.nProfiles

        self.dataOut.nCohInt = self.dataIn.nCohInt
        self.dataOut.ippSeconds = self.dataIn.ippSeconds
        self.dataOut.frequency = self.dataIn.frequency

        self.dataOut.azimuth = self.dataIn.azimuth
        self.dataOut.zenith = self.dataIn.zenith

        self.dataOut.beam.codeList = self.dataIn.beam.codeList
        self.dataOut.beam.azimuthList = self.dataIn.beam.azimuthList
        self.dataOut.beam.zenithList = self.dataIn.beam.zenithList


class selectChannels(Operation):

    def run(self, dataOut, channelList):

        channelIndexList = []
        self.dataOut = dataOut
        for channel in channelList:
            if channel not in self.dataOut.channelList:
                raise ValueError("Channel %d is not in %s" %(channel, str(self.dataOut.channelList)))

            index = self.dataOut.channelList.index(channel)
            channelIndexList.append(index)
        self.selectChannelsByIndex(channelIndexList)
        return self.dataOut

    def selectChannelsByIndex(self, channelIndexList):
        """
        Selecciona un bloque de datos en base a canales segun el channelIndexList

        Input:
            channelIndexList    :    lista sencilla de canales a seleccionar por ej. [2,3,7]

        Affected:
            self.dataOut.data
            self.dataOut.channelIndexList
            self.dataOut.nChannels
            self.dataOut.m_ProcessingHeader.totalSpectra
            self.dataOut.systemHeaderObj.numChannels
            self.dataOut.m_ProcessingHeader.blockSize

        Return:
            None
        """

        for channelIndex in channelIndexList:
            if channelIndex not in self.dataOut.channelIndexList:
                raise ValueError("The value %d in channelIndexList is not valid" %channelIndex)

        if self.dataOut.type == 'Voltage':
            if self.dataOut.flagDataAsBlock:
                """
                Si la data es obtenida por bloques, dimension = [nChannels, nProfiles, nHeis]
                """
                data = self.dataOut.data[channelIndexList,:,:]
            else:
                data = self.dataOut.data[channelIndexList,:]

            self.dataOut.data = data
            # self.dataOut.channelList = [self.dataOut.channelList[i] for i in channelIndexList]
            self.dataOut.channelList = range(len(channelIndexList))

        elif self.dataOut.type == 'Spectra':
            data_spc = self.dataOut.data_spc[channelIndexList, :]
            data_dc = self.dataOut.data_dc[channelIndexList, :]

            self.dataOut.data_spc = data_spc
            self.dataOut.data_dc = data_dc

            # self.dataOut.channelList = [self.dataOut.channelList[i] for i in channelIndexList]
            self.dataOut.channelList = range(len(channelIndexList))
            self.__selectPairsByChannel(channelIndexList)

        return 1

    def __selectPairsByChannel(self, channelList=None):

        if channelList == None:
            return

        pairsIndexListSelected = []
        for pairIndex in self.dataOut.pairsIndexList:
            # First pair
            if self.dataOut.pairsList[pairIndex][0] not in channelList:
                continue
            # Second pair
            if self.dataOut.pairsList[pairIndex][1] not in channelList:
                continue

            pairsIndexListSelected.append(pairIndex)

        if not pairsIndexListSelected:
            self.dataOut.data_cspc = None
            self.dataOut.pairsList = []
            return

        self.dataOut.data_cspc = self.dataOut.data_cspc[pairsIndexListSelected]
        self.dataOut.pairsList = [self.dataOut.pairsList[i]
                                  for i in pairsIndexListSelected]

        return

class selectHeights(Operation):

    def run(self, dataOut, minHei=None, maxHei=None):
        """
        Selecciona un bloque de datos en base a un grupo de valores de alturas segun el rango
        minHei <= height <= maxHei

        Input:
            minHei    :    valor minimo de altura a considerar
            maxHei    :    valor maximo de altura a considerar

        Affected:
            Indirectamente son cambiados varios valores a travez del metodo selectHeightsByIndex

        Return:
            1 si el metodo se ejecuto con exito caso contrario devuelve 0
        """

        self.dataOut = dataOut

        if minHei == None:
            minHei = self.dataOut.heightList[0]

        if maxHei == None:
            maxHei = self.dataOut.heightList[-1]

        if (minHei < self.dataOut.heightList[0]):
            minHei = self.dataOut.heightList[0]

        if (maxHei > self.dataOut.heightList[-1]):
            maxHei = self.dataOut.heightList[-1]

        minIndex = 0
        maxIndex = 0
        heights = self.dataOut.heightList

        inda = numpy.where(heights >= minHei)
        indb = numpy.where(heights <= maxHei)

        try:
            minIndex = inda[0][0]
        except:
            minIndex = 0

        try:
            maxIndex = indb[0][-1]
        except:
            maxIndex = len(heights)

        self.selectHeightsByIndex(minIndex, maxIndex)

        return self.dataOut

    def selectHeightsByIndex(self, minIndex, maxIndex):
        """
        Selecciona un bloque de datos en base a un grupo indices de alturas segun el rango
        minIndex <= index <= maxIndex

        Input:
            minIndex    :    valor de indice minimo de altura a considerar
            maxIndex    :    valor de indice maximo de altura a considerar

        Affected:
            self.dataOut.data
            self.dataOut.heightList

        Return:
            1 si el metodo se ejecuto con exito caso contrario devuelve 0
        """

        if self.dataOut.type == 'Voltage':
            if (minIndex < 0) or (minIndex > maxIndex):
                raise ValueError("Height index range (%d,%d) is not valid" % (minIndex, maxIndex))

            if (maxIndex >= self.dataOut.nHeights):
                maxIndex = self.dataOut.nHeights

            #voltage
            if self.dataOut.flagDataAsBlock:
                """
                Si la data es obtenida por bloques, dimension = [nChannels, nProfiles, nHeis]
                """
                data = self.dataOut.data[:,:, minIndex:maxIndex]
            else:
                data = self.dataOut.data[:, minIndex:maxIndex]

            #         firstHeight = self.dataOut.heightList[minIndex]

            self.dataOut.data = data
            self.dataOut.heightList = self.dataOut.heightList[minIndex:maxIndex]

            if self.dataOut.nHeights <= 1:
                raise ValueError("selectHeights: Too few heights. Current number of heights is %d" %(self.dataOut.nHeights))
        elif self.dataOut.type == 'Spectra':
            if (minIndex < 0) or (minIndex > maxIndex):
                raise ValueError("Error selecting heights: Index range (%d,%d) is not valid" % (
                    minIndex, maxIndex))

            if (maxIndex >= self.dataOut.nHeights):
                maxIndex = self.dataOut.nHeights - 1

            # Spectra
            data_spc = self.dataOut.data_spc[:, :, minIndex:maxIndex + 1]

            data_cspc = None
            if self.dataOut.data_cspc is not None:
                data_cspc = self.dataOut.data_cspc[:, :, minIndex:maxIndex + 1]

            data_dc = None
            if self.dataOut.data_dc is not None:
                data_dc = self.dataOut.data_dc[:, minIndex:maxIndex + 1]

            self.dataOut.data_spc = data_spc
            self.dataOut.data_cspc = data_cspc
            self.dataOut.data_dc = data_dc

            self.dataOut.heightList = self.dataOut.heightList[minIndex:maxIndex + 1]

        return 1


class filterByHeights(Operation):

    def run(self, dataOut, window):

        deltaHeight = dataOut.heightList[1] - dataOut.heightList[0]

        if window == None:
            window = (dataOut.radarControllerHeaderObj.txA/dataOut.radarControllerHeaderObj.nBaud) / deltaHeight

        newdelta = deltaHeight * window
        r = dataOut.nHeights % window
        newheights = (dataOut.nHeights-r)/window

        if newheights <= 1:
            raise ValueError("filterByHeights: Too few heights. Current number of heights is %d and window is %d" %(dataOut.nHeights, window))

        if dataOut.flagDataAsBlock:
            """
            Si la data es obtenida por bloques, dimension = [nChannels, nProfiles, nHeis]
            """
            buffer = dataOut.data[:, :, 0:int(dataOut.nHeights-r)]
            buffer = buffer.reshape(dataOut.nChannels, dataOut.nProfiles, int(dataOut.nHeights/window), window)
            buffer = numpy.sum(buffer,3)

        else:
            buffer = dataOut.data[:,0:int(dataOut.nHeights-r)]
            buffer = buffer.reshape(dataOut.nChannels,int(dataOut.nHeights/window),int(window))
            buffer = numpy.sum(buffer,2)

        dataOut.data = buffer
        dataOut.heightList = dataOut.heightList[0] + numpy.arange( newheights )*newdelta
        dataOut.windowOfFilter = window

        return dataOut


class setH0(Operation):

    def run(self, dataOut, h0, deltaHeight = None):

        if not deltaHeight:
            deltaHeight = dataOut.heightList[1] - dataOut.heightList[0]

        nHeights = dataOut.nHeights

        newHeiRange = h0 + numpy.arange(nHeights)*deltaHeight

        dataOut.heightList = newHeiRange

        return dataOut


class deFlip(Operation):

    def run(self, dataOut, channelList = []):

        data = dataOut.data.copy()

        if dataOut.flagDataAsBlock:
            flip = self.flip
            profileList = list(range(dataOut.nProfiles))

            if not channelList:
                for thisProfile in profileList:
                    data[:,thisProfile,:] = data[:,thisProfile,:]*flip
                    flip *= -1.0
            else:
                for thisChannel in channelList:
                    if thisChannel not in dataOut.channelList:
                        continue

                    for thisProfile in profileList:
                        data[thisChannel,thisProfile,:] = data[thisChannel,thisProfile,:]*flip
                        flip *= -1.0

            self.flip = flip

        else:
            if not channelList:
                data[:,:] = data[:,:]*self.flip
            else:
                for thisChannel in channelList:
                    if thisChannel not in dataOut.channelList:
                        continue

                    data[thisChannel,:] = data[thisChannel,:]*self.flip

            self.flip *= -1.

        dataOut.data = data

        return dataOut


class setAttribute(Operation):
    '''
    Set an arbitrary attribute(s) to dataOut
    '''

    def __init__(self):

        Operation.__init__(self)
        self._ready = False

    def run(self, dataOut, **kwargs):

        for key, value in kwargs.items():
            setattr(dataOut, key, value)

        return dataOut


@MPDecorator
class printAttribute(Operation):
    '''
    Print an arbitrary attribute of dataOut
    '''

    def __init__(self):

        Operation.__init__(self)

    def run(self, dataOut, attributes):

        for attr in attributes:
            if hasattr(dataOut, attr):
                log.log(getattr(dataOut, attr), attr)


class interpolateHeights(Operation):

    def run(self, dataOut, topLim, botLim):
        #69 al 72 para julia
        #82-84 para meteoros
        if len(numpy.shape(dataOut.data))==2:
            sampInterp = (dataOut.data[:,botLim-1] + dataOut.data[:,topLim+1])/2
            sampInterp = numpy.transpose(numpy.tile(sampInterp,(topLim-botLim + 1,1)))
            #dataOut.data[:,botLim:limSup+1] = sampInterp
            dataOut.data[:,botLim:topLim+1] = sampInterp
        else:
            nHeights = dataOut.data.shape[2]
            x = numpy.hstack((numpy.arange(botLim),numpy.arange(topLim+1,nHeights)))
            y = dataOut.data[:,:,list(range(botLim))+list(range(topLim+1,nHeights))]
            f = interpolate.interp1d(x, y, axis = 2)
            xnew = numpy.arange(botLim,topLim+1)
            ynew = f(xnew)
            dataOut.data[:,:,botLim:topLim+1]  = ynew

        return dataOut


class CohInt(Operation):

    isConfig = False
    __profIndex = 0
    __byTime = False
    __initime = None
    __lastdatatime = None
    __integrationtime = None
    __buffer = None
    __bufferStride = []
    __dataReady = False
    __profIndexStride = 0
    __dataToPutStride = False
    n = None

    def __init__(self, **kwargs):

        Operation.__init__(self, **kwargs)

    def setup(self, n=None, timeInterval=None, stride=None, overlapping=False, byblock=False):
        """
        Set the parameters of the integration class.

        Inputs:

            n               :    Number of coherent integrations
            timeInterval    :    Time of integration. If the parameter "n" is selected this one does not work
            overlapping     :
        """

        self.__initime = None
        self.__lastdatatime = 0
        self.__buffer = None
        self.__dataReady = False
        self.byblock = byblock
        self.stride = stride

        if n == None and timeInterval == None:
            raise ValueError("n or timeInterval should be specified ...")

        if n != None:
            self.n = n
            self.__byTime = False
        else:
            self.__integrationtime = timeInterval #* 60. #if (type(timeInterval)!=integer) -> change this line
            self.n = 9999
            self.__byTime = True

        if overlapping:
            self.__withOverlapping = True
            self.__buffer = None
        else:
            self.__withOverlapping = False
            self.__buffer = 0

        self.__profIndex = 0

    def putData(self, data):

        """
        Add a profile to the __buffer and increase in one the __profileIndex

        """

        if not self.__withOverlapping:
            self.__buffer += data.copy()
            self.__profIndex += 1
            return

        #Overlapping data
        nChannels, nHeis = data.shape
        data = numpy.reshape(data, (1, nChannels, nHeis))

        #If the buffer is empty then it takes the data value
        if self.__buffer is None:
            self.__buffer = data
            self.__profIndex += 1
            return

        #If the buffer length is lower than n then stakcing the data value
        if self.__profIndex < self.n:
            self.__buffer = numpy.vstack((self.__buffer, data))
            self.__profIndex += 1
            return

        #If the buffer length is equal to n then replacing the last buffer value with the data value
        self.__buffer = numpy.roll(self.__buffer, -1, axis=0)
        self.__buffer[self.n-1] = data
        self.__profIndex = self.n
        return


    def pushData(self):
        """
        Return the sum of the last profiles and the profiles used in the sum.

        Affected:

        self.__profileIndex

        """

        if not self.__withOverlapping:
            data = self.__buffer
            n = self.__profIndex

            self.__buffer = 0
            self.__profIndex = 0

            return data, n

        #Integration with Overlapping
        data = numpy.sum(self.__buffer, axis=0)
        # print data
        # raise
        n = self.__profIndex

        return data, n

    def byProfiles(self, data):

        self.__dataReady = False
        avgdata = None
        #         n = None
        # print data
        # raise
        self.putData(data)

        if self.__profIndex == self.n:
            avgdata, n = self.pushData()
            self.__dataReady = True

        return avgdata

    def byTime(self, data, datatime):

        self.__dataReady = False
        avgdata = None
        n = None

        self.putData(data)

        if (datatime - self.__initime) >= self.__integrationtime:
            avgdata, n = self.pushData()
            self.n = n
            self.__dataReady = True

        return avgdata

    def integrateByStride(self, data, datatime):
        # print data
        if self.__profIndex == 0:
            self.__buffer = [[data.copy(), datatime]]
        else:
            self.__buffer.append([data.copy(),datatime])
        self.__profIndex += 1
        self.__dataReady = False

        if self.__profIndex == self.n * self.stride :
            self.__dataToPutStride = True
            self.__profIndexStride = 0
            self.__profIndex = 0
            self.__bufferStride = []
            for i in range(self.stride):
                current = self.__buffer[i::self.stride]
                data = numpy.sum([t[0] for t in current], axis=0)
                avgdatatime = numpy.average([t[1] for t in current])
                # print data
                self.__bufferStride.append((data, avgdatatime))

        if self.__dataToPutStride:
            self.__dataReady = True
            self.__profIndexStride += 1
            if self.__profIndexStride == self.stride:
                self.__dataToPutStride = False
            # print self.__bufferStride[self.__profIndexStride - 1]
            # raise
            return self.__bufferStride[self.__profIndexStride - 1]


        return None, None

    def integrate(self, data, datatime=None):

        if self.__initime == None:
            self.__initime = datatime

        if self.__byTime:
            avgdata = self.byTime(data, datatime)
        else:
            avgdata = self.byProfiles(data)


        self.__lastdatatime = datatime

        if avgdata is None:
            return None, None

        avgdatatime = self.__initime

        deltatime = datatime - self.__lastdatatime

        if not self.__withOverlapping:
            self.__initime = datatime
        else:
            self.__initime += deltatime

        return avgdata, avgdatatime

    def integrateByBlock(self, dataOut):

        times = int(dataOut.data.shape[1]/self.n)
        avgdata = numpy.zeros((dataOut.nChannels, times, dataOut.nHeights), dtype=numpy.complex)

        id_min = 0
        id_max = self.n

        for i in range(times):
            junk = dataOut.data[:,id_min:id_max,:]
            avgdata[:,i,:] = junk.sum(axis=1)
            id_min += self.n
            id_max += self.n

        timeInterval = dataOut.ippSeconds*self.n
        avgdatatime = (times - 1) * timeInterval + dataOut.utctime
        self.__dataReady = True
        return avgdata, avgdatatime

    def run(self, dataOut, n=None, timeInterval=None, stride=None, overlapping=False, byblock=False, **kwargs):

        if not self.isConfig:
            self.setup(n=n, stride=stride, timeInterval=timeInterval, overlapping=overlapping, byblock=byblock, **kwargs)
            self.isConfig = True

        if dataOut.flagDataAsBlock:
            """
            Si la data es leida por bloques, dimension = [nChannels, nProfiles, nHeis]
            """
            avgdata, avgdatatime = self.integrateByBlock(dataOut)
            dataOut.nProfiles /= self.n
        else:
            if stride is None:
                avgdata, avgdatatime = self.integrate(dataOut.data, dataOut.utctime)
            else:
                avgdata, avgdatatime = self.integrateByStride(dataOut.data, dataOut.utctime)


        #   dataOut.timeInterval *= n
        dataOut.flagNoData = True

        if self.__dataReady:
            dataOut.data = avgdata
            if not dataOut.flagCohInt:
                dataOut.nCohInt *= self.n
                dataOut.flagCohInt = True
            dataOut.utctime = avgdatatime
            # print avgdata, avgdatatime
            # raise
            #   dataOut.timeInterval = dataOut.ippSeconds * dataOut.nCohInt
            dataOut.flagNoData = False
        return dataOut

class Decoder(Operation):

    isConfig = False
    __profIndex = 0

    code = None

    nCode = None
    nBaud = None

    def __init__(self, **kwargs):

        Operation.__init__(self, **kwargs)

        self.times = None
        self.osamp = None
    #         self.__setValues = False
        self.isConfig = False
        self.setupReq = False
    def setup(self, code, osamp, dataOut):

        self.__profIndex = 0

        self.code = code

        self.nCode = len(code)
        self.nBaud = len(code[0])

        if (osamp != None) and (osamp >1):
            self.osamp = osamp
            self.code = numpy.repeat(code, repeats=self.osamp, axis=1)
            self.nBaud = self.nBaud*self.osamp

        self.__nChannels = dataOut.nChannels
        self.__nProfiles = dataOut.nProfiles
        self.__nHeis = dataOut.nHeights

        if self.__nHeis < self.nBaud:
            raise ValueError('Number of heights (%d) should be greater than number of bauds (%d)' %(self.__nHeis, self.nBaud))

        #Frequency
        __codeBuffer = numpy.zeros((self.nCode, self.__nHeis), dtype=numpy.complex)

        __codeBuffer[:,0:self.nBaud] = self.code

        self.fft_code = numpy.conj(numpy.fft.fft(__codeBuffer, axis=1))

        if dataOut.flagDataAsBlock:

            self.ndatadec = self.__nHeis #- self.nBaud + 1

            self.datadecTime = numpy.zeros((self.__nChannels, self.__nProfiles, self.ndatadec), dtype=numpy.complex)

        else:

            #Time
            self.ndatadec = self.__nHeis #- self.nBaud + 1

            self.datadecTime = numpy.zeros((self.__nChannels, self.ndatadec), dtype=numpy.complex)

    def __convolutionInFreq(self, data):

        fft_code = self.fft_code[self.__profIndex].reshape(1,-1)

        fft_data = numpy.fft.fft(data, axis=1)

        conv = fft_data*fft_code

        data = numpy.fft.ifft(conv,axis=1)

        return data

    def __convolutionInFreqOpt(self, data):

        raise NotImplementedError

    def __convolutionInTime(self, data):

        code = self.code[self.__profIndex]
        for i in range(self.__nChannels):
            self.datadecTime[i,:] = numpy.correlate(data[i,:], code, mode='full')[self.nBaud-1:]

        return self.datadecTime

    def __convolutionByBlockInTime(self, data):

        repetitions = int(self.__nProfiles / self.nCode)
        junk = numpy.lib.stride_tricks.as_strided(self.code, (repetitions, self.code.size), (0, self.code.itemsize))
        junk = junk.flatten()
        code_block = numpy.reshape(junk, (self.nCode*repetitions, self.nBaud))
        profilesList = range(self.__nProfiles)

        for i in range(self.__nChannels):
            for j in profilesList:
                self.datadecTime[i,j,:] = numpy.correlate(data[i,j,:], code_block[j,:], mode='full')[self.nBaud-1:]
        return self.datadecTime

    def __convolutionByBlockInFreq(self, data):

        raise NotImplementedError("Decoder by frequency fro Blocks not implemented")


        fft_code = self.fft_code[self.__profIndex].reshape(1,-1)

        fft_data = numpy.fft.fft(data, axis=2)

        conv = fft_data*fft_code

        data = numpy.fft.ifft(conv,axis=2)

        return data


    def run(self, dataOut, code=None, nCode=None, nBaud=None, mode = 0, osamp=None, times=None):

        if dataOut.flagDecodeData:
            print("This data is already decoded, recoding again ...")

        if not self.isConfig:

            if code is None:
                if dataOut.code is None:
                    raise ValueError("Code could not be read from %s instance. Enter a value in Code parameter" %dataOut.type)

                code = dataOut.code
            else:
                code = numpy.array(code).reshape(nCode,nBaud)
            self.setup(code, osamp, dataOut)

            self.isConfig = True

            if mode == 3:
                sys.stderr.write("Decoder Warning: mode=%d is not valid, using mode=0\n" %mode)

            if times != None:
                sys.stderr.write("Decoder Warning: Argument 'times' in not used anymore\n")

        if self.code is None:
            print("Fail decoding: Code is not defined.")
            return

        self.__nProfiles = dataOut.nProfiles
        datadec = None

        if mode == 3:
            mode = 0

        if dataOut.flagDataAsBlock:
            """
            Decoding when data have been read as block,
            """

            if mode == 0:
                datadec = self.__convolutionByBlockInTime(dataOut.data)
            if mode == 1:
                datadec = self.__convolutionByBlockInFreq(dataOut.data)
        else:
            """
            Decoding when data have been read profile by profile
            """
            if mode == 0:
                datadec = self.__convolutionInTime(dataOut.data)

            if mode == 1:
                datadec = self.__convolutionInFreq(dataOut.data)

            if mode == 2:
                datadec = self.__convolutionInFreqOpt(dataOut.data)

        if datadec is None:
            raise ValueError("Codification mode selected is not valid: mode=%d. Try selecting 0 or 1" %mode)

        dataOut.code = self.code
        dataOut.nCode = self.nCode
        dataOut.nBaud = self.nBaud

        dataOut.data = datadec

        dataOut.heightList = dataOut.heightList[0:datadec.shape[-1]]

        dataOut.flagDecodeData = True #asumo q la data esta decodificada

        if self.__profIndex == self.nCode-1:
            self.__profIndex = 0
            return dataOut

        self.__profIndex += 1

        return dataOut
    #        dataOut.flagDeflipData = True #asumo q la data no esta sin flip


class ProfileConcat(Operation):

    isConfig = False
    buffer = None

    def __init__(self, **kwargs):

        Operation.__init__(self, **kwargs)
        self.profileIndex = 0

    def reset(self):
        self.buffer = numpy.zeros_like(self.buffer)
        self.start_index = 0
        self.times = 1

    def setup(self, data, m, n=1):
        self.buffer = numpy.zeros((data.shape[0],data.shape[1]*m),dtype=type(data[0,0]))
        self.nHeights = data.shape[1]#.nHeights
        self.start_index = 0
        self.times = 1

    def concat(self, data):

        self.buffer[:,self.start_index:self.nHeights*self.times] = data.copy()
        self.start_index = self.start_index + self.nHeights

    def run(self, dataOut, m):
        dataOut.flagNoData = True

        if not self.isConfig:
            self.setup(dataOut.data, m, 1)
            self.isConfig = True

        if dataOut.flagDataAsBlock:
            raise ValueError("ProfileConcat can only be used when voltage have been read profile by profile, getBlock = False")

        else:
            self.concat(dataOut.data)
            self.times += 1
            if self.times > m:
                dataOut.data = self.buffer
                self.reset()
                dataOut.flagNoData = False
                # se deben actualizar mas propiedades del header y del objeto dataOut, por ejemplo, las alturas
                deltaHeight = dataOut.heightList[1] - dataOut.heightList[0]
                xf = dataOut.heightList[0] + dataOut.nHeights * deltaHeight * m
                dataOut.heightList = numpy.arange(dataOut.heightList[0], xf, deltaHeight)
                dataOut.ippSeconds *= m
        return dataOut

class ProfileSelector(Operation):

    profileIndex = None
    # Tamanho total de los perfiles
    nProfiles = None

    def __init__(self, **kwargs):

        Operation.__init__(self, **kwargs)
        self.profileIndex = 0

    def incProfileIndex(self):

        self.profileIndex += 1

        if self.profileIndex >= self.nProfiles:
            self.profileIndex = 0

    def isThisProfileInRange(self, profileIndex, minIndex, maxIndex):

        if profileIndex < minIndex:
            return False

        if profileIndex > maxIndex:
            return False

        return True

    def isThisProfileInList(self, profileIndex, profileList):

        if profileIndex not in profileList:
            return False

        return True

    def run(self, dataOut, profileList=None, profileRangeList=None, beam=None, byblock=False, rangeList = None, nProfiles=None):

        """
        ProfileSelector:

        Inputs:
            profileList        :    Index of profiles selected. Example: profileList = (0,1,2,7,8)

            profileRangeList    :    Minimum and maximum profile indexes. Example: profileRangeList = (4, 30)

            rangeList            :    List of profile ranges. Example: rangeList = ((4, 30), (32, 64), (128, 256))

        """

        if rangeList is not None:
            if type(rangeList[0]) not in (tuple, list):
                rangeList = [rangeList]

        dataOut.flagNoData = True

        if dataOut.flagDataAsBlock:
            """
            data dimension  = [nChannels, nProfiles, nHeis]
            """
            if profileList != None:
                dataOut.data = dataOut.data[:,profileList,:]

            if profileRangeList != None:
                minIndex = profileRangeList[0]
                maxIndex = profileRangeList[1]
                profileList = list(range(minIndex, maxIndex+1))

                dataOut.data = dataOut.data[:,minIndex:maxIndex+1,:]

            if rangeList != None:

                profileList = []

                for thisRange in rangeList:
                    minIndex = thisRange[0]
                    maxIndex = thisRange[1]

                    profileList.extend(list(range(minIndex, maxIndex+1)))

                dataOut.data = dataOut.data[:,profileList,:]

            dataOut.nProfiles = len(profileList)
            dataOut.profileIndex = dataOut.nProfiles - 1
            dataOut.flagNoData = False

            return dataOut

        """
        data dimension  = [nChannels, nHeis]
        """

        if profileList != None:

            if self.isThisProfileInList(dataOut.profileIndex, profileList):

                self.nProfiles = len(profileList)
                dataOut.nProfiles = self.nProfiles
                dataOut.profileIndex = self.profileIndex
                dataOut.flagNoData = False

                self.incProfileIndex()
            return dataOut

        if profileRangeList != None:

            minIndex = profileRangeList[0]
            maxIndex = profileRangeList[1]

            if self.isThisProfileInRange(dataOut.profileIndex, minIndex, maxIndex):

                self.nProfiles = maxIndex - minIndex + 1
                dataOut.nProfiles = self.nProfiles
                dataOut.profileIndex = self.profileIndex
                dataOut.flagNoData = False

                self.incProfileIndex()
            return dataOut

        if rangeList != None:

            nProfiles = 0

            for thisRange in rangeList:
                minIndex = thisRange[0]
                maxIndex = thisRange[1]

                nProfiles += maxIndex - minIndex + 1

            for thisRange in rangeList:

                minIndex = thisRange[0]
                maxIndex = thisRange[1]

                if self.isThisProfileInRange(dataOut.profileIndex, minIndex, maxIndex):

                    self.nProfiles = nProfiles
                    dataOut.nProfiles = self.nProfiles
                    dataOut.profileIndex = self.profileIndex
                    dataOut.flagNoData = False

                    self.incProfileIndex()

                    break

            return dataOut


        if beam != None: #beam is only for AMISR data
            if self.isThisProfileInList(dataOut.profileIndex, dataOut.beamRangeDict[beam]):
                dataOut.flagNoData = False
                dataOut.profileIndex = self.profileIndex

                self.incProfileIndex()

            return dataOut

        raise ValueError("ProfileSelector needs profileList, profileRangeList or rangeList parameter")


class Reshaper(Operation):

    def __init__(self, **kwargs):

        Operation.__init__(self, **kwargs)

        self.__buffer = None
        self.__nitems = 0

    def __appendProfile(self, dataOut, nTxs):

        if self.__buffer is None:
            shape = (dataOut.nChannels, int(dataOut.nHeights/nTxs) )
            self.__buffer = numpy.empty(shape, dtype = dataOut.data.dtype)

        ini = dataOut.nHeights * self.__nitems
        end = ini + dataOut.nHeights

        self.__buffer[:, ini:end] = dataOut.data

        self.__nitems += 1

        return int(self.__nitems*nTxs)

    def __getBuffer(self):

        if self.__nitems == int(1./self.__nTxs):

            self.__nitems = 0

            return self.__buffer.copy()

        return None

    def __checkInputs(self, dataOut, shape, nTxs):

        if shape is None and nTxs is None:
            raise ValueError("Reshaper: shape of factor should be defined")

        if nTxs:
            if nTxs < 0:
                raise ValueError("nTxs should be greater than 0")

            if nTxs < 1 and dataOut.nProfiles % (1./nTxs) != 0:
                raise ValueError("nProfiles= %d is not divisibled by (1./nTxs) = %f" %(dataOut.nProfiles, (1./nTxs)))

            shape = [dataOut.nChannels, dataOut.nProfiles*nTxs, dataOut.nHeights/nTxs]

            return shape, nTxs

        if len(shape) != 2 and len(shape) !=  3:
            raise ValueError("shape dimension should be equal to 2 or 3. shape = (nProfiles, nHeis) or (nChannels, nProfiles, nHeis). Actually shape = (%d, %d, %d)" %(dataOut.nChannels, dataOut.nProfiles, dataOut.nHeights))

        if len(shape) == 2:
            shape_tuple = [dataOut.nChannels]
            shape_tuple.extend(shape)
        else:
            shape_tuple = list(shape)

        nTxs = 1.0*shape_tuple[1]/dataOut.nProfiles

        return shape_tuple, nTxs

    def run(self, dataOut, shape=None, nTxs=None):

        shape_tuple, self.__nTxs = self.__checkInputs(dataOut, shape, nTxs)

        dataOut.flagNoData = True
        profileIndex = None

        if dataOut.flagDataAsBlock:

            dataOut.data = numpy.reshape(dataOut.data, shape_tuple)
            dataOut.flagNoData = False

            profileIndex = int(dataOut.nProfiles*self.__nTxs) - 1

        else:

            if self.__nTxs < 1:

                self.__appendProfile(dataOut, self.__nTxs)
                new_data = self.__getBuffer()

                if new_data is not None:
                    dataOut.data = new_data
                    dataOut.flagNoData = False

                    profileIndex = dataOut.profileIndex*nTxs

            else:
                raise ValueError("nTxs should be greater than 0 and lower than 1, or use VoltageReader(..., getblock=True)")

        deltaHeight = dataOut.heightList[1] - dataOut.heightList[0]

        dataOut.heightList = numpy.arange(dataOut.nHeights/self.__nTxs) * deltaHeight + dataOut.heightList[0]

        dataOut.nProfiles = int(dataOut.nProfiles*self.__nTxs)

        dataOut.profileIndex = profileIndex

        dataOut.ippSeconds /= self.__nTxs

        return dataOut

class SplitProfiles(Operation):

    def __init__(self, **kwargs):

        Operation.__init__(self, **kwargs)

    def run(self, dataOut, n):

        dataOut.flagNoData = True
        profileIndex = None

        if dataOut.flagDataAsBlock:

            #nchannels, nprofiles, nsamples
            shape = dataOut.data.shape

            if shape[2] % n != 0:
                raise ValueError("Could not split the data, n=%d has to be multiple of %d" %(n, shape[2]))

            new_shape = shape[0], shape[1]*n, int(shape[2]/n)

            dataOut.data = numpy.reshape(dataOut.data, new_shape)
            dataOut.flagNoData = False

            profileIndex = int(dataOut.nProfiles/n) - 1

        else:

            raise ValueError("Could not split the data when is read Profile by Profile. Use VoltageReader(..., getblock=True)")

        deltaHeight = dataOut.heightList[1] - dataOut.heightList[0]

        dataOut.heightList = numpy.arange(dataOut.nHeights/n) * deltaHeight + dataOut.heightList[0]

        dataOut.nProfiles = int(dataOut.nProfiles*n)

        dataOut.profileIndex = profileIndex

        dataOut.ippSeconds /= n

        return dataOut

class CombineProfiles(Operation):
    def __init__(self, **kwargs):

        Operation.__init__(self, **kwargs)

        self.__remData = None
        self.__profileIndex = 0

    def run(self, dataOut, n):

        dataOut.flagNoData = True
        profileIndex = None

        if dataOut.flagDataAsBlock:

            #nchannels, nprofiles, nsamples
            shape = dataOut.data.shape
            new_shape = shape[0], shape[1]/n, shape[2]*n

            if shape[1] % n != 0:
                raise ValueError("Could not split the data, n=%d has to be multiple of %d" %(n, shape[1]))

            dataOut.data = numpy.reshape(dataOut.data, new_shape)
            dataOut.flagNoData = False

            profileIndex = int(dataOut.nProfiles*n) - 1

        else:

            #nchannels, nsamples
            if self.__remData is None:
                newData = dataOut.data
            else:
                newData = numpy.concatenate((self.__remData, dataOut.data), axis=1)

            self.__profileIndex += 1

            if self.__profileIndex < n:
                self.__remData = newData
                #continue
                return

            self.__profileIndex = 0
            self.__remData = None

            dataOut.data = newData
            dataOut.flagNoData = False

            profileIndex = dataOut.profileIndex/n


        deltaHeight = dataOut.heightList[1] - dataOut.heightList[0]

        dataOut.heightList = numpy.arange(dataOut.nHeights*n) * deltaHeight + dataOut.heightList[0]

        dataOut.nProfiles = int(dataOut.nProfiles/n)

        dataOut.profileIndex = profileIndex

        dataOut.ippSeconds *= n

        return dataOut

class PulsePairVoltage(Operation):
    '''
    Function PulsePair(Signal Power, Velocity)
    The real component of Lag[0] provides Intensity Information
    The imag component of Lag[1] Phase provides Velocity Information

    Configuration Parameters:
    nPRF = Number of Several PRF
    theta = Degree Azimuth angel Boundaries

    Input:
          self.dataOut
          lag[N]
    Affected:
          self.dataOut.spc
    '''
    isConfig       = False
    __profIndex    = 0
    __initime      = None
    __lastdatatime = None
    __buffer       = None
    noise          = None
    __dataReady    = False
    n              = None
    __nch          = 0
    __nHeis        = 0
    removeDC       = False
    ipp            = None
    lambda_        = 0

    def __init__(self,**kwargs):
        Operation.__init__(self,**kwargs)

    def setup(self, dataOut, n = None, removeDC=False):
        '''
        n= Numero de PRF's de entrada
        '''
        self.__initime        = None
        self.__lastdatatime   = 0
        self.__dataReady      = False
        self.__buffer         = 0
        self.__profIndex      = 0
        self.noise            = None
        self.__nch            = dataOut.nChannels
        self.__nHeis          = dataOut.nHeights
        self.removeDC         = removeDC
        self.lambda_          = 3.0e8/(9345.0e6)
        self.ippSec           = dataOut.ippSeconds
        self.nCohInt          = dataOut.nCohInt
        print("IPPseconds",dataOut.ippSeconds)

        print("ELVALOR DE n es:", n)
        if n == None:
            raise ValueError("n should be specified.")

        if n != None:
            if n<2:
                raise ValueError("n should be greater than 2")

        self.n       = n
        self.__nProf = n

        self.__buffer = numpy.zeros((dataOut.nChannels,
                                           n,
                                           dataOut.nHeights),
                                          dtype='complex')

    def putData(self,data):
        '''
        Add a profile to he __buffer and increase in one the __profiel Index
        '''
        self.__buffer[:,self.__profIndex,:]= data
        self.__profIndex      += 1
        return

    def pushData(self,dataOut):
        '''
        Return the PULSEPAIR and the profiles used in the operation
        Affected :  self.__profileIndex
        '''
        #················· Remove DC···································
        if self.removeDC==True:
            mean    = numpy.mean(self.__buffer,1)
            tmp     = mean.reshape(self.__nch,1,self.__nHeis)
            dc= numpy.tile(tmp,[1,self.__nProf,1])
            self.__buffer = self.__buffer -  dc
        #··················Calculo de Potencia ························
        pair0       = self.__buffer*numpy.conj(self.__buffer)
        pair0       = pair0.real
        lag_0       = numpy.sum(pair0,1)
        #··················Calculo de Ruido x canal····················
        self.noise  = numpy.zeros(self.__nch)
        for i in range(self.__nch):
            daux         = numpy.sort(pair0[i,:,:],axis= None)
            self.noise[i]=hildebrand_sekhon( daux ,self.nCohInt)

        self.noise       = self.noise.reshape(self.__nch,1)
        self.noise       = numpy.tile(self.noise,[1,self.__nHeis])
        noise_buffer     = self.noise.reshape(self.__nch,1,self.__nHeis)
        noise_buffer     = numpy.tile(noise_buffer,[1,self.__nProf,1])
        #·················· Potencia recibida= P , Potencia senal = S , Ruido= N··
        #··················   P= S+N  ,P=lag_0/N ·································
        #···················· Power ··················································
        data_power       = lag_0/(self.n*self.nCohInt)
        #------------------  Senal  ···················································
        data_intensity   = pair0 - noise_buffer
        data_intensity   = numpy.sum(data_intensity,axis=1)*(self.n*self.nCohInt)#*self.nCohInt)
        #data_intensity   = (lag_0-self.noise*self.n)*(self.n*self.nCohInt)
        for i in range(self.__nch):
            for j in range(self.__nHeis):
                if data_intensity[i][j]  < 0:
                    data_intensity[i][j] = numpy.min(numpy.absolute(data_intensity[i][j]))

        #················· Calculo de Frecuencia y Velocidad doppler········
        pair1            = self.__buffer[:,:-1,:]*numpy.conjugate(self.__buffer[:,1:,:])
        lag_1            = numpy.sum(pair1,1)
        data_freq        = (-1/(2.0*math.pi*self.ippSec*self.nCohInt))*numpy.angle(lag_1)
        data_velocity    = (self.lambda_/2.0)*data_freq

        #················ Potencia promedio estimada de la Senal···········
        lag_0            = lag_0/self.n
        S                = lag_0-self.noise

        #················ Frecuencia Doppler promedio ·····················
        lag_1            = lag_1/(self.n-1)
        R1               = numpy.abs(lag_1)

        #················ Calculo del SNR··································
        data_snrPP       = S/self.noise
        for i in range(self.__nch):
            for j in range(self.__nHeis):
                if data_snrPP[i][j]  < 1.e-20:
                    data_snrPP[i][j] = 1.e-20

        #················· Calculo del ancho espectral ······················
        L                = S/R1
        L                = numpy.where(L<0,1,L)
        L                = numpy.log(L)
        tmp              = numpy.sqrt(numpy.absolute(L))
        data_specwidth   = (self.lambda_/(2*math.sqrt(2)*math.pi*self.ippSec*self.nCohInt))*tmp*numpy.sign(L)
        n                = self.__profIndex

        self.__buffer    = numpy.zeros((self.__nch, self.__nProf,self.__nHeis),  dtype='complex')
        self.__profIndex = 0
        return data_power,data_intensity,data_velocity,data_snrPP,data_specwidth,n


    def pulsePairbyProfiles(self,dataOut):

        self.__dataReady     =  False
        data_power           =  None
        data_intensity       =  None
        data_velocity        =  None
        data_specwidth       =  None
        data_snrPP           =  None
        self.putData(data=dataOut.data)
        if self.__profIndex  == self.n:
            data_power,data_intensity, data_velocity,data_snrPP,data_specwidth, n   = self.pushData(dataOut=dataOut)
            self.__dataReady                   = True

        return data_power, data_intensity, data_velocity, data_snrPP, data_specwidth


    def pulsePairOp(self, dataOut, datatime= None):

        if self.__initime == None:
            self.__initime = datatime
        data_power, data_intensity, data_velocity, data_snrPP, data_specwidth = self.pulsePairbyProfiles(dataOut)
        self.__lastdatatime           = datatime

        if data_power is None:
            return None, None, None,None,None,None

        avgdatatime    = self.__initime
        deltatime      = datatime - self.__lastdatatime
        self.__initime = datatime

        return data_power, data_intensity, data_velocity, data_snrPP, data_specwidth, avgdatatime

    def run(self, dataOut,n = None,removeDC= False, overlapping= False,**kwargs):

        if not self.isConfig:
            self.setup(dataOut = dataOut, n    = n , removeDC=removeDC , **kwargs)
            self.isConfig   = True
        data_power, data_intensity, data_velocity,data_snrPP,data_specwidth, avgdatatime = self.pulsePairOp(dataOut, dataOut.utctime)
        dataOut.flagNoData                         = True

        if self.__dataReady:
            dataOut.nCohInt        *= self.n
            dataOut.dataPP_POW      = data_intensity # S
            dataOut.dataPP_POWER    = data_power     # P
            dataOut.dataPP_DOP      = data_velocity
            dataOut.dataPP_SNR      = data_snrPP
            dataOut.dataPP_WIDTH    = data_specwidth
            dataOut.PRFbyAngle      = self.n         #numero de PRF*cada angulo rotado que equivale a un tiempo.
            dataOut.utctime         = avgdatatime
            dataOut.flagNoData      = False
        return dataOut



# import collections
# from scipy.stats import mode
#
# class Synchronize(Operation):
#
#     isConfig = False
#     __profIndex = 0
#
#     def __init__(self, **kwargs):
#
#         Operation.__init__(self, **kwargs)
# #         self.isConfig = False
#         self.__powBuffer = None
#         self.__startIndex = 0
#         self.__pulseFound = False
#
#     def __findTxPulse(self, dataOut, channel=0, pulse_with = None):
#
#         #Read data
#
#         powerdB = dataOut.getPower(channel = channel)
#         noisedB = dataOut.getNoise(channel = channel)[0]
#
#         self.__powBuffer.extend(powerdB.flatten())
#
#         dataArray = numpy.array(self.__powBuffer)
#
#         filteredPower = numpy.correlate(dataArray, dataArray[0:self.__nSamples], "same")
#
#         maxValue = numpy.nanmax(filteredPower)
#
#         if maxValue < noisedB + 10:
#             #No se encuentra ningun pulso de transmision
#             return None
#
#         maxValuesIndex = numpy.where(filteredPower > maxValue - 0.1*abs(maxValue))[0]
#
#         if len(maxValuesIndex) < 2:
#             #Solo se encontro un solo pulso de transmision de un baudio, esperando por el siguiente TX
#             return None
#
#         phasedMaxValuesIndex = maxValuesIndex - self.__nSamples
#
#         #Seleccionar solo valores con un espaciamiento de nSamples
#         pulseIndex = numpy.intersect1d(maxValuesIndex, phasedMaxValuesIndex)
#
#         if len(pulseIndex) < 2:
#             #Solo se encontro un pulso de transmision con ancho mayor a 1
#             return None
#
#         spacing = pulseIndex[1:] - pulseIndex[:-1]
#
#         #remover senales que se distancien menos de 10 unidades o muestras
#         #(No deberian existir IPP menor a 10 unidades)
#
#         realIndex = numpy.where(spacing > 10 )[0]
#
#         if len(realIndex) < 2:
#             #Solo se encontro un pulso de transmision con ancho mayor a 1
#             return None
#
#         #Eliminar pulsos anchos (deja solo la diferencia entre IPPs)
#         realPulseIndex = pulseIndex[realIndex]
#
#         period = mode(realPulseIndex[1:] - realPulseIndex[:-1])[0][0]
#
#         print "IPP = %d samples" %period
#
#         self.__newNSamples = dataOut.nHeights #int(period)
#         self.__startIndex = int(realPulseIndex[0])
#
#         return 1
#
#
#     def setup(self, nSamples, nChannels, buffer_size = 4):
#
#         self.__powBuffer = collections.deque(numpy.zeros( buffer_size*nSamples,dtype=numpy.float),
#                                           maxlen = buffer_size*nSamples)
#
#         bufferList = []
#
#         for i in range(nChannels):
#             bufferByChannel = collections.deque(numpy.zeros( buffer_size*nSamples, dtype=numpy.complex) +  numpy.NAN,
#                                           maxlen = buffer_size*nSamples)
#
#             bufferList.append(bufferByChannel)
#
#         self.__nSamples = nSamples
#         self.__nChannels = nChannels
#         self.__bufferList = bufferList
#
#     def run(self, dataOut, channel = 0):
#
#         if not self.isConfig:
#             nSamples = dataOut.nHeights
#             nChannels = dataOut.nChannels
#             self.setup(nSamples, nChannels)
#             self.isConfig = True
#
#         #Append new data to internal buffer
#         for thisChannel in range(self.__nChannels):
#             bufferByChannel = self.__bufferList[thisChannel]
#             bufferByChannel.extend(dataOut.data[thisChannel])
#
#         if self.__pulseFound:
#             self.__startIndex -= self.__nSamples
#
#         #Finding Tx Pulse
#         if not self.__pulseFound:
#             indexFound = self.__findTxPulse(dataOut, channel)
#
#             if indexFound == None:
#                 dataOut.flagNoData = True
#                 return
#
#             self.__arrayBuffer = numpy.zeros((self.__nChannels, self.__newNSamples), dtype = numpy.complex)
#             self.__pulseFound = True
#             self.__startIndex = indexFound
#
#         #If pulse was found ...
#         for thisChannel in range(self.__nChannels):
#             bufferByChannel = self.__bufferList[thisChannel]
#             #print self.__startIndex
#             x = numpy.array(bufferByChannel)
#             self.__arrayBuffer[thisChannel] = x[self.__startIndex:self.__startIndex+self.__newNSamples]
#
#         deltaHeight = dataOut.heightList[1] - dataOut.heightList[0]
#         dataOut.heightList = numpy.arange(self.__newNSamples)*deltaHeight
# #             dataOut.ippSeconds = (self.__newNSamples / deltaHeight)/1e6
#
#         dataOut.data = self.__arrayBuffer
#
#         self.__startIndex += self.__newNSamples
#
#         return



'''
Created on Jul 3, 2014

@author: roj-com0419
'''

import os,sys
import time,datetime
import h5py
import numpy
import fnmatch
import re

from schainpy.model.data.jroheaderIO import RadarControllerHeader, SystemHeader
from schainpy.model.data.jrodata import Voltage
from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation


def isNumber(str):
    """
    Chequea si el conjunto de caracteres que componen un string puede ser convertidos a un numero.

    Excepciones:
        Si un determinado string no puede ser convertido a numero
    Input:
        str, string al cual se le analiza para determinar si convertible a un numero o no

    Return:
        True    :    si el string es uno numerico
        False   :    no es un string numerico
    """
    try:
        float( str )
        return True
    except:
        return False

def getFileFromSet(path, ext, set=None):
    validFilelist = []
    fileList = os.listdir(path)


    if len(fileList) < 1:
        return None

    # 0 1234 567 89A BCDE
    # H YYYY DDD SSS .ext

    for thisFile in fileList:
        try:
            number= int(thisFile[6:16])

     #       year = int(thisFile[1:5])
     #       doy  = int(thisFile[5:8])
        except:
            continue

        if (os.path.splitext(thisFile)[-1].lower() != ext.lower()):
            continue

        validFilelist.append(thisFile)

    if len(validFilelist) < 1:
        return None

    validFilelist = sorted( validFilelist, key=str.lower )

    if set == None:
        return validFilelist[-1]

    print("set =" ,set)
    for thisFile in validFilelist:
        if set <= int(thisFile[6:16]):
            print(thisFile,int(thisFile[6:16]))
            return thisFile

    return validFilelist[-1]

    myfile = fnmatch.filter(validFilelist,'*%10d*'%(set))
    #myfile = fnmatch.filter(validFilelist,'*%4.4d%3.3d%3.3d*'%(year,doy,set))

    if len(myfile)!= 0:
        return myfile[0]
    else:
        filename = '*%10.10d%s'%(set,ext.lower())
        print('the filename %s does not exist'%filename)
        print('...going to the last file: ')

    if validFilelist:
        validFilelist = sorted( validFilelist, key=str.lower )
        return validFilelist[-1]

    return None

def getlastFileFromPath(path, ext):
    """
Depura el fileList dejando solo los que cumplan el formato de "res-xxxxxx.ext"
    al final de la depuracion devuelve el ultimo file de la lista que quedo.

    Input:
        fileList    :    lista conteniendo todos los files (sin path) que componen una determinada carpeta
        ext         :    extension de los files contenidos en una carpeta

    Return:
        El ultimo file de una determinada carpeta, no se considera el path.
    """
    validFilelist = []
    fileList = os.listdir(path)

    # 0 1234 567 89A BCDE
    # H YYYY DDD SSS .ext

    for thisFile in fileList:

        try:
            number= int(thisFile[6:16])
        except:
            print("There is a file or folder with different format")
        if not isNumber(number):
            continue

#        year = thisFile[1:5]
#        if not isNumber(year):
#            continue

#       doy = thisFile[5:8]
#        if not isNumber(doy):
#            continue

        number= int(number)
#        year = int(year)
#        doy = int(doy)

        if (os.path.splitext(thisFile)[-1].lower() != ext.lower()):
            continue


        validFilelist.append(thisFile)


    if validFilelist:
        validFilelist = sorted( validFilelist, key=str.lower )
        return validFilelist[-1]

    return None



class HFReader(ProcessingUnit):
    '''
    classdocs
    '''
    path     = None
    startDate= None
    endDate  = None
    startTime= None
    endTime  = None
    walk     = None
    isConfig = False
    dataOut=None
    nTries = 3
    ext      = ".hdf5"

    def __init__(self, **kwargs):
        '''
        Constructor
        '''
        ProcessingUnit.__init__(self, **kwargs)

        self.isConfig =False

        self.datablock = None

        self.filename_current=None

        self.utc = 0

        self.ext='.hdf5'

        self.flagIsNewFile = 1

        #-------------------------------------------------
        self.fileIndex=None

        self.profileIndex_offset=None

        self.filenameList=[]

        self.hfFilePointer= None

        self.filename_online = None

        self.status=True

        self.flagNoMoreFiles= False

        self.__waitForNewFile = 20


        #--------------------------------------------------

        self.dataOut = self.createObjByDefault()


    def createObjByDefault(self):

        dataObj = Voltage()

        return dataObj

    def setObjProperties(self):

        pass

    def getBlockDimension(self):
        """
        Obtiene la cantidad de puntos a leer por cada bloque de datos

        Affected:
            self.blocksize

        Return:
            None
        """
        pts2read =self.nChannels*self.nHeights*self.nProfiles
        self.blocksize = pts2read

    def __readHeader(self):

        self.nProfiles = 100
        self.nHeights = 1000
        self.nChannels = 2
        self.__firstHeigth=0
        self.__nSamples=1000
        self.__deltaHeigth=1.5
        self.__sample_rate=1e5
        #self.__frequency=2.72e6
        #self.__frequency=3.64e6
        self.__frequency=None
        self.__online = False
        self.filename_next_set=None

        #print "Frequency of Operation:", self.__frequency


    def __setParameters(self,path='', startDate='',endDate='',startTime='', endTime='', walk=''):
        self.path = path
        self.startDate = startDate
        self.endDate = endDate
        self.startTime = startTime
        self.endTime = endTime
        self.walk = walk

    def __checkPath(self):
        if os.path.exists(self.path):
            self.status=1
        else:
            self.status=0
            print('Path %s does not exits'%self.path)
            return
        return

    def __selDates(self, hf_dirname_format):
        try:
            dir_hf_filename= self.path+"/"+hf_dirname_format
            fp= h5py.File(dir_hf_filename,'r')
            hipoc=fp['t'].value
            fp.close()
            date_time=datetime.datetime.utcfromtimestamp(hipoc)
            year =int(date_time[0:4])
            month=int(date_time[5:7])
            dom  =int(date_time[8:10])
            thisDate= datetime.date(year,month,dom)
            if (thisDate>=self.startDate and thisDate <= self.endDate):
                return hf_dirname_format
        except:
            return None

    def __findDataForDates(self,online=False):
        if not(self.status):
            return None

        pat = '\d+.\d+'
        dirnameList = [re.search(pat,x) for x in os.listdir(self.path)]
        dirnameList = [x for x in dirnameList if x!=None]
        dirnameList = [x.string for x in dirnameList]
        if not(online):

            dirnameList = [self.__selDates(x) for x in dirnameList]
            dirnameList = [x for x in dirnameList if x!=None]

        if len(dirnameList)>0:
                self.status = 1
                self.dirnameList = dirnameList
                self.dirnameList.sort()

        else:
            self.status = 0
            return None

    def __getTimeFromData(self):
        startDateTime_Reader = datetime.datetime.combine(self.startDate,self.startTime)
        endDateTime_Reader = datetime.datetime.combine(self.endDate,self.endTime)
        print('Filtering Files from %s to %s'%(startDateTime_Reader, endDateTime_Reader))
        print('........................................')
        filter_filenameList=[]
        self.filenameList.sort()
        for i in range(len(self.filenameList)-1):
            filename=self.filenameList[i]
            dir_hf_filename= filename
            fp= h5py.File(dir_hf_filename,'r')
            hipoc=fp['t'].value
            hipoc=hipoc+self.timezone
            date_time=datetime.datetime.utcfromtimestamp(hipoc)
            fp.close()
            year =int(date_time[0:4])
            month=int(date_time[5:7])
            dom  =int(date_time[8:10])
            hour =int(date_time[11:13])
            min  =int(date_time[14:16])
            sec  =int(date_time[17:19])
            this_time=datetime.datetime(year,month,dom,hour,min,sec)
            if (this_time>=startDateTime_Reader and this_time <= endDateTime_Reader):
                filter_filenameList.append(filename)
        filter_filenameList.sort()
        self.filenameList = filter_filenameList
        return 1

    def __getFilenameList(self):
        #print "hola"
        #print self.dirnameList
        dirList = [os.path.join(self.path,x) for x in self.dirnameList]
        self.filenameList= dirList
        #print self.filenameList
        #print "pase",len(self.filenameList)

    def __selectDataForTimes(self, online=False):

        if not(self.status):
            return None
        #----------------
        self.__getFilenameList()
        #----------------
        if not(online):
            if not(self.all):
                self.__getTimeFromData()
            if len(self.filenameList)>0:
                self.status=1
                self.filenameList.sort()
            else:
                self.status=0
                return None
        else:
            if self.set != None:

                filename=getFileFromSet(self.path,self.ext,self.set)

                if self.flag_nextfile==True:
                    self.dirnameList=[filename]
                    fullfilename=self.path+"/"+filename
                    self.filenameList=[fullfilename]
                    self.filename_next_set=int(filename[6:16])+10

                    self.flag_nextfile=False
                else:
                    print(filename)
                    print("PRIMERA CONDICION")
                    #if self.filename_next_set== int(filename[6:16]):
                    print("TODO BIEN")

                    if filename == None:
                        raise ValueError("corregir")

                    self.dirnameList=[filename]
                    fullfilename=self.path+"/"+filename
                    self.filenameList=[fullfilename]
                    self.filename_next_set=int(filename[6:16])+10
                    print("Setting next file",self.filename_next_set)
                    self.set=int(filename[6:16])
                    if True:
                        pass
                    else:
                        print("ESTOY AQUI PORQUE NO EXISTE EL SIGUIENTE ARCHIVO")

            else:
                filename =getlastFileFromPath(self.path,self.ext)

                if self.flag_nextfile==True:
                    self.dirnameList=[filename]
                    fullfilename=self.path+"/"+filename
                    self.filenameList=[self.filenameList[-1]]
                    self.filename_next_set=int(filename[6:16])+10

                    self.flag_nextfile=False
                else:
                   filename=getFileFromSet(self.path,self.ext,self.set)
                   print(filename)
                   print("PRIMERA CONDICION")
                    #if self.filename_next_set== int(filename[6:16]):
                   print("TODO BIEN")

                   if filename == None:
                       raise ValueError("corregir")

                   self.dirnameList=[filename]
                   fullfilename=self.path+"/"+filename
                   self.filenameList=[fullfilename]
                   self.filename_next_set=int(filename[6:16])+10
                   print("Setting next file",self.filename_next_set)
                   self.set=int(filename[6:16])
                   if True:
                       pass
                   else:
                       print("ESTOY AQUI PORQUE NO EXISTE EL SIGUIENTE ARCHIVO")



    def searchFilesOffLine(self,
                            path,
                            startDate,
                            endDate,
                            ext,
                            startTime=datetime.time(0,0,0),
                            endTime=datetime.time(23,59,59),
                            walk=True):

        self.__setParameters(path, startDate, endDate, startTime, endTime, walk)

        self.__checkPath()

        self.__findDataForDates()
        #print self.dirnameList

        self.__selectDataForTimes()

        for i in range(len(self.filenameList)):
            print("%s"% (self.filenameList[i]))

        return

    def searchFilesOnLine(self,
                            path,
                            expLabel= "",
                            ext=None,
                            startDate=None,
                            endDate=None,
                            walk=True,
                            set=None):


        startDate = datetime.datetime.utcnow().date()
        endDate = datetime.datetime.utcnow().date()

        self.__setParameters(path=path,startDate=startDate,endDate=endDate,walk=walk)

        self.__checkPath()

        fullpath=path
        print("%s folder was found: " %(fullpath ))

        if set == None:
            self.set=None
            filename =getlastFileFromPath(fullpath,ext)
            startDate= datetime.datetime.utcnow().date
            endDate= datetime.datetime.utcnow().date()
#
        else:
            filename= getFileFromSet(fullpath,ext,set)
            startDate=None
            endDate=None
#
        if not (filename):
            return None,None,None,None,None
        #print "%s file was found" %(filename)

#
#         dir_hf_filename= self.path+"/"+filename
#         fp= h5py.File(dir_hf_filename,'r')
#         hipoc=fp['t'].value
#         fp.close()
#         date_time=datetime.datetime.utcfromtimestamp(hipoc)
#
#         year =int(date_time[0:4])
#         month=int(date_time[5:7])
#         dom  =int(date_time[8:10])
#         set= int(filename[4:10])
#         self.set=set-1
        #self.dirnameList=[filename]
        filenameList= fullpath+"/"+filename
        self.dirnameList=[filename]
        self.filenameList=[filenameList]
        self.flag_nextfile=True

        #self.__findDataForDates(online=True)
        #self.dirnameList=[self.dirnameList[-1]]
        #print self.dirnameList
        #self.__selectDataForTimes(online=True)
        #return fullpath,filename,year,month,dom,set
        return

    def __setNextFile(self,online=False):
        """
        """
        if not(online):
            newFile = self.__setNextFileOffline()
        else:
            newFile = self.__setNextFileOnline()

        if not(newFile):
            return 0
        return 1

    def __setNextFileOffline(self):
        """
        """
        idFile= self.fileIndex
        while(True):
            idFile += 1
            if not (idFile < len(self.filenameList)):
                self.flagNoMoreFiles = 1
                print("No more Files")
                return 0
            filename = self.filenameList[idFile]
            hfFilePointer =h5py.File(filename,'r')

            epoc=hfFilePointer['t'].value
            #this_time=datetime.datetime(year,month,dom,hour,min,sec)
            break

        self.flagIsNewFile = 1
        self.fileIndex = idFile
        self.filename = filename

        self.hfFilePointer = hfFilePointer
        hfFilePointer.close()
        self.__t0=epoc
        print("Setting the file: %s"%self.filename)

        return 1

    def __setNextFileOnline(self):
        """
        """
        print("SOY NONE",self.set)
        if self.set==None:
            pass
        else:
            self.set +=10

        filename = self.filenameList[0]#fullfilename
        if self.filename_online != None:
            self.__selectDataForTimes(online=True)
            filename = self.filenameList[0]
            while self.filename_online == filename:
                print('waiting %d seconds to get a new file...'%(self.__waitForNewFile))
                time.sleep(self.__waitForNewFile)
                #self.__findDataForDates(online=True)
                self.set=self.filename_next_set
                self.__selectDataForTimes(online=True)
                filename = self.filenameList[0]
                sizeoffile=os.path.getsize(filename)

        #print filename
        sizeoffile=os.path.getsize(filename)
        if sizeoffile<1670240:
            print("%s is not the rigth  size"%filename)
            delay=50
            print('waiting %d seconds for delay...'%(delay))
            time.sleep(delay)
        sizeoffile=os.path.getsize(filename)
        if sizeoffile<1670240:
            delay=50
            print('waiting %d  more seconds for delay...'%(delay))
            time.sleep(delay)

        sizeoffile=os.path.getsize(filename)
        if sizeoffile<1670240:
            delay=50
            print('waiting %d  more seconds for delay...'%(delay))
            time.sleep(delay)

        try:
            hfFilePointer=h5py.File(filename,'r')

        except:
            print("Error reading file %s"%filename)

        self.filename_online=filename
        epoc=hfFilePointer['t'].value

        self.hfFilePointer=hfFilePointer
        hfFilePointer.close()
        self.__t0=epoc


        self.flagIsNewFile = 1
        self.filename = filename

        print("Setting the file: %s"%self.filename)
        return 1

    def __getExpParameters(self):
        if not(self.status):
            return None

    def setup(self,
               path = None,
               startDate = None,
               endDate = None,
               startTime = datetime.time(0,0,0),
               endTime = datetime.time(23,59,59),
               set = None,
               expLabel = "",
               ext = None,
               all=0,
               timezone=0,
               online = False,
               delay = 60,
               walk = True):
        '''
        In this method we should set all initial parameters.

        '''
        if path==None:
            raise ValueError("The path is not valid")

        if ext==None:
            ext = self.ext

        self.timezone= timezone
        self.online= online
        self.all=all
        #if set==None:

        #print set
        if not(online):
            print("Searching files in offline mode...")

            self.searchFilesOffLine(path, startDate, endDate, ext, startTime, endTime, walk)
        else:
            print("Searching files in online mode...")
            self.searchFilesOnLine(path, walk,ext,set=set)
            if set==None:
                pass
            else:
                self.set=set-10

#             for nTries in range(self.nTries):
#
#                 fullpath,file,year,month,day,set = self.searchFilesOnLine(path=path,expLabel=expLabel,ext=ext, walk=walk,set=set)
#
#                 if fullpath:
#                     break
#                 print '\tWaiting %0.2f sec for an valid file in %s: try %02d ...' % (self.delay, path, nTries+1)
#                 time.sleep(self.delay)
#             if not(fullpath):
#                 print "There ins't valid files in %s" % path
#                 return None


        if not(self.filenameList):
            print("There  is no files into the folder: %s"%(path))
            sys.exit(-1)

        self.__getExpParameters()


        self.fileIndex = -1

        self.__setNextFile(online)

        self.__readMetadata()

        self.__setLocalVariables()

        self.__setHeaderDO()
        #self.profileIndex_offset= 0

        #self.profileIndex = self.profileIndex_offset

        self.isConfig = True

    def __readMetadata(self):
        self.__readHeader()


    def __setLocalVariables(self):

        self.datablock = numpy.zeros((self.nChannels, self.nHeights,self.nProfiles), dtype = numpy.complex)
        #



        self.profileIndex = 9999


    def __setHeaderDO(self):


        self.dataOut.radarControllerHeaderObj = RadarControllerHeader()

        self.dataOut.systemHeaderObj = SystemHeader()


        #---------------------------------------------------------
        self.dataOut.systemHeaderObj.nProfiles=100
        self.dataOut.systemHeaderObj.nSamples=1000


        SAMPLING_STRUCTURE=[('h0', '<f4'), ('dh', '<f4'), ('nsa', '<u4')]
        self.dataOut.radarControllerHeaderObj.samplingWindow=numpy.zeros((1,),SAMPLING_STRUCTURE)
        self.dataOut.radarControllerHeaderObj.samplingWindow['h0']=0
        self.dataOut.radarControllerHeaderObj.samplingWindow['dh']=1.5
        self.dataOut.radarControllerHeaderObj.samplingWindow['nsa']=1000
        self.dataOut.radarControllerHeaderObj.nHeights=int(self.dataOut.radarControllerHeaderObj.samplingWindow['nsa'])
        self.dataOut.radarControllerHeaderObj.firstHeight = self.dataOut.radarControllerHeaderObj.samplingWindow['h0']
        self.dataOut.radarControllerHeaderObj.deltaHeight = self.dataOut.radarControllerHeaderObj.samplingWindow['dh']
        self.dataOut.radarControllerHeaderObj.samplesWin = self.dataOut.radarControllerHeaderObj.samplingWindow['nsa']

        self.dataOut.radarControllerHeaderObj.nWindows=1
        self.dataOut.radarControllerHeaderObj.codetype=0
        self.dataOut.radarControllerHeaderObj.numTaus=0
        #self.dataOut.radarControllerHeaderObj.Taus = numpy.zeros((1,),'<f4')


        #self.dataOut.radarControllerHeaderObj.nCode=numpy.zeros((1,), '<u4')
        #self.dataOut.radarControllerHeaderObj.nBaud=numpy.zeros((1,), '<u4')
        #self.dataOut.radarControllerHeaderObj.code=numpy.zeros(0)

        self.dataOut.radarControllerHeaderObj.code_size=0
        self.dataOut.nBaud=0
        self.dataOut.nCode=0
        self.dataOut.nPairs=0


        #---------------------------------------------------------

        self.dataOut.type = "Voltage"

        self.dataOut.data = None

        self.dataOut.dtype = numpy.dtype([('real','<f4'),('imag','<f4')])

        self.dataOut.nProfiles = 1

        self.dataOut.heightList = self.__firstHeigth + numpy.arange(self.__nSamples, dtype = numpy.float)*self.__deltaHeigth

        self.dataOut.channelList = list(range(self.nChannels))

        #self.dataOut.channelIndexList = None

        self.dataOut.flagNoData = True

        #Set to TRUE if the data is discontinuous
        self.dataOut.flagDiscontinuousBlock = False

        self.dataOut.utctime = None

        self.dataOut.timeZone = self.timezone

        self.dataOut.dstFlag = 0

        self.dataOut.errorCount = 0

        self.dataOut.nCohInt = 1

        self.dataOut.blocksize = self.dataOut.getNChannels() * self.dataOut.getNHeights()

        self.dataOut.flagDecodeData = False #asumo que la data esta decodificada

        self.dataOut.flagDeflipData = False #asumo que la data esta sin flip

        self.dataOut.flagShiftFFT = False

        self.dataOut.ippSeconds = 1.0*self.__nSamples/self.__sample_rate

        #Time interval between profiles
        #self.dataOut.timeInterval =self.dataOut.ippSeconds * self.dataOut.nCohInt


        self.dataOut.frequency = self.__frequency

        self.dataOut.realtime = self.__online

    def __hasNotDataInBuffer(self):

        if self.profileIndex >= self.nProfiles:
            return 1

        return 0

    def readNextBlock(self):
        if not(self.__setNewBlock()):
            return 0

        if not(self.readBlock()):
            return 0

        return 1

    def __setNewBlock(self):

        if self.hfFilePointer==None:
            return 0

        if self.flagIsNewFile:
            return 1

        if self.profileIndex < self.nProfiles:
            return 1

        self.__setNextFile(self.online)

        return 1



    def readBlock(self):
        fp=h5py.File(self.filename,'r')
                      #Puntero que apunta al archivo hdf5
        ch0=(fp['ch0']).value    #Primer canal (100,1000)--(perfiles,alturas)
        ch1=(fp['ch1']).value    #Segundo canal (100,1000)--(perfiles,alturas)
        fp.close()
        ch0= ch0.swapaxes(0,1)   #Primer canal (100,1000)--(alturas,perfiles)
        ch1= ch1.swapaxes(0,1)   #Segundo canal (100,1000)--(alturas,perfiles)
        self.datablock = numpy.array([ch0,ch1])
        self.flagIsNewFile=0

        self.profileIndex=0

        return 1

    def getData(self):
        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            return 0

        if self.__hasNotDataInBuffer():
            if not(self.readNextBlock()):
                self.dataOut.flagNodata=True
                return 0

        ##############################
        ##############################
        self.dataOut.data = self.datablock[:,:,self.profileIndex]
        self.dataOut.utctime = self.__t0 + self.dataOut.ippSeconds*self.profileIndex
        self.dataOut.profileIndex= self.profileIndex
        self.dataOut.flagNoData=False
        self.profileIndex +=1

        return self.dataOut.data


    def run(self, **kwargs):
        '''
        This method will be called many times so here you should put all your code
        '''

        if not self.isConfig:
            self.setup(**kwargs)
            self.isConfig = True
        self.getData()


'''
Created on Jul 15, 2014

@author: roj-idl71
'''
import time
import threading
import pickle

# try:
#     from gevent import sleep
# except:
from time import sleep

SERIALIZER = cPickle

# from schainpy.serializer import DynamicSerializer
from schainpy.model.io.jroIO_usrp import USRPReader
from schainpy.model.serializer.data import obj2Serial

class USRPReaderAPI(USRPReader, threading.Thread):
    
#     __isBufferEmpty = True
    
    __DATAKEYLIST = ['data','utctime','flagNoData']
    
    def __init__(self, serializer='msgpack'):
        
        threading.Thread.__init__(self)
        USRPReader.__init__(self)
        
#         self.__serializerObj = DynamicSerializer.DynamicSerializer('msgpack')
        self.__mySerial = None
        self.__isBufferEmpty = True
        
        self.setSerializer(serializer)
    
    def setSerializer(self, serializer):
        
        self.__serializer = serializer
    
    def getSerializer(self):
        
        return self.__serializer
        
    def getProfileIndex(self):
        
        return self.profileIndex
    
    def getSerialMetaData(self):
        
        if self.__isBufferEmpty:
            ini = time.time()
            
            while True:
                
                if not self.__isBufferEmpty:
                    break
                
                if time.time() - ini > 20:
                    break
                
                sleep(1e-12)
                    
            
#             if not self.getData():
#                 self.__isBufferEmpty = False
#                 return None
    
        if self.dataOut.flagNoData:
            return None
        
        myMetadataSerial = obj2Serial(self.dataOut,
                                      serializer = self.__serializer)
        
        return myMetadataSerial
    
    def getSerialData(self):
        
        if self.__isBufferEmpty:
            ini = time.time()
            
            while True:
                
                if not self.__isBufferEmpty:
                    break
                
                if time.time() - ini > 20:
                    break
                
                sleep(1e-12)
                    
            
#             if not self.getData():
#                 self.__isBufferEmpty = False
#                 return None
    
        if self.dataOut.flagNoData:
            return None
        
        self.__isBufferEmpty = True
        
        return self.__mySerial
    
    def run(self):
         
        '''
        This method will be called once when start() is called
        '''
         
        if not self.isConfig:
            raise RuntimeError('setup() method has to be called before start()')
        
        print("Running ...")
        
        while True:
             
            if not self.__isBufferEmpty:
                sleep(1e-12)
                continue
             
            if not self.getData():
                break
            
            print(".", end=' ')
            
            self.__mySerial = obj2Serial(self.dataOut,
                                         keyList = self.__DATAKEYLIST,
                                         serializer = self.__serializer)
            self.__isBufferEmpty = False
             
#             print self.profileIndex
#             print 'wait 1 second'
            
#             sleep(0.1)

        print("Closing thread")
        
        return


# Copyright (c) 2012-2020 Jicamarca Radio Observatory
# All rights reserved.
#
# Distributed under the terms of the BSD 3-clause license.
"""schainpy is an open source library to read, write and process radar data

Signal Chain is a radar data processing library wich includes modules to read,
and write different files formats, besides modules to process and visualize the
data.
"""

import os
from setuptools import setup, Extension
from setuptools.command.build_ext import build_ext as _build_ext
from schainpy import __version__

DOCLINES = __doc__.split("\n")

class build_ext(_build_ext):
    def finalize_options(self):
        _build_ext.finalize_options(self)
        # Prevent numpy from thinking it is still in its setup process:
        __builtins__.__NUMPY_SETUP__ = False
        import numpy
        self.include_dirs.append(numpy.get_include())

setup(
    name = "schainpy",
    version = __version__,
    description = DOCLINES[0],
    long_description = "\n".join(DOCLINES[2:]),
    url = "https://github.com/JRO-Peru/schain",
    author = "Jicamarca Radio Observatory",
    author_email = "jro-developers@igp.gob.pe",
    license="BSD-3-Clause",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering",
    ],
    packages = {
        'schainpy',
        'schainpy.model',
        'schainpy.model.data',
        'schainpy.model.graphics',
        'schainpy.model.io',
        'schainpy.model.proc',
        'schainpy.model.utils',
        'schainpy.utils',
        'schainpy.gui',
        'schainpy.cli',
        },
    package_data = {'': ['schain.conf.template'],
                    'schainpy.files': ['*.oga']
                    },
    include_package_data = False,
    scripts = ['schainpy/gui/schainGUI'],
    entry_points = {
        'console_scripts': [
            'schain = schainpy.cli.cli:main',
            ],
        },
    cmdclass = {'build_ext': build_ext},
    ext_modules=[
        Extension("schainpy.model.data._noise", ["schainc/_noise.c"]),
        ],
    setup_requires = ["numpy"],
    install_requires = [
        "scipy",
        "h5py",
        "matplotlib",
        "pyzmq",
        "fuzzywuzzy",
        "click",
        ],
)



'''
@author: Daniel Suarez
'''

import os
import sys
import glob
import fnmatch
import datetime
import time
import re
import h5py
import numpy

from schainpy.model.proc.jroproc_base import ProcessingUnit, Operation
from schainpy.model.data.jroamisr import AMISR

try:
    from gevent import sleep
except:
    from time import sleep
    
class RadacHeader():
    def __init__(self, fp):
        header = 'Raw11/Data/RadacHeader'
        self.beamCodeByPulse = fp.get(header+'/BeamCode')
        self.beamCode = fp.get('Raw11/Data/Beamcodes')
        self.code = fp.get(header+'/Code')
        self.frameCount = fp.get(header+'/FrameCount')
        self.modeGroup = fp.get(header+'/ModeGroup')
        self.nsamplesPulse = fp.get(header+'/NSamplesPulse')
        self.pulseCount = fp.get(header+'/PulseCount')
        self.radacTime = fp.get(header+'/RadacTime')
        self.timeCount = fp.get(header+'/TimeCount')
        self.timeStatus = fp.get(header+'/TimeStatus')
        
        self.nrecords = self.pulseCount.shape[0] #nblocks
        self.npulses = self.pulseCount.shape[1] #nprofile
        self.nsamples = self.nsamplesPulse[0,0] #ngates
        self.nbeams = self.beamCode.shape[1]
        
    
    def getIndexRangeToPulse(self, idrecord=0):
        #indexToZero = numpy.where(self.pulseCount.value[idrecord,:]==0)
        #startPulseCountId = indexToZero[0][0]
        #endPulseCountId = startPulseCountId - 1
        #range1 = numpy.arange(startPulseCountId,self.npulses,1)
        #range2 = numpy.arange(0,startPulseCountId,1)
        #return range1, range2
        zero = 0
        npulse = max(self.pulseCount[0,:]+1)-1
        looking_index = numpy.where(self.pulseCount.value[idrecord,:]==npulse)[0]
        getLastIndex = looking_index[-1]
        index_data = numpy.arange(0,getLastIndex+1,1)
        index_buffer = numpy.arange(getLastIndex+1,self.npulses,1)
        return index_data, index_buffer
    
class AMISRReader(ProcessingUnit):
    
    path = None
    startDate = None
    endDate = None
    startTime = None
    endTime = None
    walk = None
    isConfig = False
    
    def __init__(self):
        self.set = None
        self.subset = None
        self.extension_file = '.h5'
        self.dtc_str = 'dtc'
        self.dtc_id = 0
        self.status = True
        self.isConfig = False
        self.dirnameList = []
        self.filenameList = []
        self.fileIndex = None
        self.flagNoMoreFiles = False
        self.flagIsNewFile = 0
        self.filename = ''
        self.amisrFilePointer = None
        self.radacHeaderObj = None
        self.dataOut = self.__createObjByDefault()
        self.datablock = None
        self.rest_datablock = None
        self.range = None
        self.idrecord_count = 0
        self.profileIndex = 0
        self.index_amisr_sample = None 
        self.index_amisr_buffer = None
        self.beamCodeByFrame = None
        self.radacTimeByFrame = None
        #atributos originales tal y como esta en el archivo de datos
        self.beamCodesFromFile = None
        self.radacTimeFromFile = None
        self.rangeFromFile = None
        self.dataByFrame = None
        self.dataset = None
        
        self.beamCodeDict = {}
        self.beamRangeDict = {}
        
        #experiment cgf file
        self.npulsesint_fromfile = None
        self.recordsperfile_fromfile = None
        self.nbeamcodes_fromfile = None
        self.ngates_fromfile = None
        self.ippSeconds_fromfile = None
        self.frequency_h5file = None
        
        
        self.__firstFile = True
        self.buffer_radactime = None
        
        self.index4_schain_datablock = None
        self.index4_buffer = None
        self.schain_datablock = None
        self.buffer = None
        self.linear_pulseCount = None
        self.npulseByFrame = None
        self.profileIndex_offset = None
        self.timezone = 'ut'
        
        self.__waitForNewFile = 20
        self.__filename_online = None 
        
    def __createObjByDefault(self):
        
        dataObj = AMISR()
        
        return dataObj
    
    def __setParameters(self,path='', startDate='',endDate='',startTime='', endTime='', walk=''):
        self.path = path
        self.startDate = startDate
        self.endDate = endDate
        self.startTime = startTime
        self.endTime = endTime
        self.walk = walk
    
    def __checkPath(self):
        if os.path.exists(self.path):
            self.status = 1
        else:
            self.status = 0
            print('Path:%s does not exists'%self.path)
            
        return
    
    def __selDates(self, amisr_dirname_format):
        try:
            year = int(amisr_dirname_format[0:4])
            month = int(amisr_dirname_format[4:6])
            dom = int(amisr_dirname_format[6:8])
            thisDate = datetime.date(year,month,dom)
            
            if (thisDate>=self.startDate and thisDate <= self.endDate):
                return amisr_dirname_format
        except:
            return None
    
    def __findDataForDates(self,online=False):
        
        
        
        if not(self.status):
            return None
        
        pat = '\d+.\d+'
        dirnameList = [re.search(pat,x) for x in os.listdir(self.path)]
        dirnameList = [x for x in dirnameList if x!=None]
        dirnameList = [x.string for x in dirnameList]
        if not(online):
            dirnameList = [self.__selDates(x) for x in dirnameList]
            dirnameList = [x for x in dirnameList if x!=None]
        if len(dirnameList)>0:
            self.status = 1
            self.dirnameList = dirnameList
            self.dirnameList.sort()
        else:
            self.status = 0
            return None
    
    def __getTimeFromData(self):
        startDateTime_Reader = datetime.datetime.combine(self.startDate,self.startTime)
        endDateTime_Reader = datetime.datetime.combine(self.endDate,self.endTime)

        print('Filtering Files from %s to %s'%(startDateTime_Reader, endDateTime_Reader))
        print('........................................')
        filter_filenameList = []
        self.filenameList.sort()
        for i in range(len(self.filenameList)-1):
            filename = self.filenameList[i]
            fp = h5py.File(filename,'r')
            time_str = fp.get('Time/RadacTimeString')
            
            startDateTimeStr_File = time_str[0][0].split('.')[0]
            junk = time.strptime(startDateTimeStr_File, '%Y-%m-%d %H:%M:%S')
            startDateTime_File = datetime.datetime(junk.tm_year,junk.tm_mon,junk.tm_mday,junk.tm_hour, junk.tm_min, junk.tm_sec)
            
            endDateTimeStr_File = time_str[-1][-1].split('.')[0]
            junk = time.strptime(endDateTimeStr_File, '%Y-%m-%d %H:%M:%S')
            endDateTime_File = datetime.datetime(junk.tm_year,junk.tm_mon,junk.tm_mday,junk.tm_hour, junk.tm_min, junk.tm_sec)
            
            fp.close()
            
            if self.timezone == 'lt':
                startDateTime_File = startDateTime_File - datetime.timedelta(minutes = 300)
                endDateTime_File = endDateTime_File - datetime.timedelta(minutes = 300)

            if (endDateTime_File>=startDateTime_Reader and endDateTime_File<endDateTime_Reader):
                #self.filenameList.remove(filename)
                filter_filenameList.append(filename)
        
        filter_filenameList.sort()
        self.filenameList = filter_filenameList
        return 1
    
    def __filterByGlob1(self, dirName):
        filter_files = glob.glob1(dirName, '*.*%s'%self.extension_file)
        filterDict = {}
        filterDict.setdefault(dirName)
        filterDict[dirName] = filter_files
        return filterDict
    
    def __getFilenameList(self, fileListInKeys, dirList):
        for value in fileListInKeys:
            dirName = list(value.keys())[0]
            for file in value[dirName]:
                filename = os.path.join(dirName, file)
                self.filenameList.append(filename)
    
    
    def __selectDataForTimes(self, online=False):
        #aun no esta implementado el filtro for tiempo
        if not(self.status):
            return None
        
        dirList = [os.path.join(self.path,x) for x in self.dirnameList]
        
        fileListInKeys = [self.__filterByGlob1(x) for x in dirList]
        
        self.__getFilenameList(fileListInKeys, dirList)
        if not(online):
            #filtro por tiempo
            if not(self.all):
                self.__getTimeFromData()

            if len(self.filenameList)>0:
                self.status = 1
                self.filenameList.sort()
            else:
                self.status = 0
                return None
            
        else:
            #get the last file - 1
            self.filenameList = [self.filenameList[-2]]
        
        new_dirnameList = []
        for dirname in self.dirnameList:
            junk = numpy.array([dirname in x for x in self.filenameList])
            junk_sum = junk.sum()
            if junk_sum > 0:
                new_dirnameList.append(dirname)
        self.dirnameList = new_dirnameList
        return 1
        
    def searchFilesOnLine(self,
                            path,
                            walk=True):
        
        startDate = datetime.datetime.utcnow().date()
        endDate = datetime.datetime.utcnow().date()
        
        self.__setParameters(path=path, startDate=startDate, endDate=endDate, walk=walk)
        
        self.__checkPath()
        
        self.__findDataForDates(online=True)
        
        self.dirnameList = [self.dirnameList[-1]]
        
        self.__selectDataForTimes(online=True)
        
        return
        
    
    def searchFilesOffLine(self,
                            path,
                            startDate,
                            endDate,
                            startTime=datetime.time(0,0,0),
                            endTime=datetime.time(23,59,59),
                            walk=True):
        
        self.__setParameters(path, startDate, endDate, startTime, endTime, walk)
        
        self.__checkPath()
        
        self.__findDataForDates()
        
        self.__selectDataForTimes()
        
        for i in range(len(self.filenameList)):
            print("%s" %(self.filenameList[i]))
        
        return

    def __setNextFileOffline(self):
        idFile = self.fileIndex

        while (True):
            idFile += 1
            if not(idFile < len(self.filenameList)):
                self.flagNoMoreFiles = 1
                print("No more Files")
                return 0

            filename = self.filenameList[idFile]

            amisrFilePointer = h5py.File(filename,'r')
            
            break

        self.flagIsNewFile = 1
        self.fileIndex = idFile
        self.filename = filename

        self.amisrFilePointer = amisrFilePointer

        print("Setting the file: %s"%self.filename)

        return 1
    
    
    def __setNextFileOnline(self):
        filename = self.filenameList[0]
        if self.__filename_online != None:
            self.__selectDataForTimes(online=True)
            filename = self.filenameList[0]
            while self.__filename_online == filename:
                print('waiting %d seconds to get a new file...'%(self.__waitForNewFile))
                sleep(self.__waitForNewFile)
                self.__selectDataForTimes(online=True)
                filename = self.filenameList[0]
        
        self.__filename_online = filename
        
        self.amisrFilePointer = h5py.File(filename,'r')
        self.flagIsNewFile = 1
        self.filename = filename
        print("Setting the file: %s"%self.filename)
        return 1
    
    
    def __readHeader(self):
        self.radacHeaderObj = RadacHeader(self.amisrFilePointer)
        
        #update values from experiment cfg file
        if self.radacHeaderObj.nrecords == self.recordsperfile_fromfile:
            self.radacHeaderObj.nrecords = self.recordsperfile_fromfile
        self.radacHeaderObj.nbeams = self.nbeamcodes_fromfile
        self.radacHeaderObj.npulses = self.npulsesint_fromfile
        self.radacHeaderObj.nsamples = self.ngates_fromfile
        
        #looking index list for data
        start_index = self.radacHeaderObj.pulseCount[0,:][0]
        end_index = self.radacHeaderObj.npulses
        range4data = list(range(start_index, end_index))
        self.index4_schain_datablock = numpy.array(range4data)
        
        buffer_start_index = 0
        buffer_end_index = self.radacHeaderObj.pulseCount[0,:][0]
        range4buffer = list(range(buffer_start_index, buffer_end_index))
        self.index4_buffer = numpy.array(range4buffer)
        
        self.linear_pulseCount = numpy.array(range4data + range4buffer)
        self.npulseByFrame = max(self.radacHeaderObj.pulseCount[0,:]+1)

        #get tuning frequency
        frequency_h5file_dataset = self.amisrFilePointer.get('Rx'+'/TuningFrequency')
        self.frequency_h5file = frequency_h5file_dataset[0,0]
        
        self.flagIsNewFile = 1
    
    def __getBeamCode(self):
        self.beamCodeDict = {}
        self.beamRangeDict = {}
        
        beamCodeMap = self.amisrFilePointer.get('Setup/BeamcodeMap')
        
        for i in range(len(self.radacHeaderObj.beamCode[0,:])):
            self.beamCodeDict.setdefault(i)
            self.beamRangeDict.setdefault(i)
            beamcodeValue = self.radacHeaderObj.beamCode[0,i]
            beamcodeIndex = numpy.where(beamCodeMap[:,0] == beamcodeValue)[0][0]
            x = beamCodeMap[beamcodeIndex][1]
            y = beamCodeMap[beamcodeIndex][2]
            z = beamCodeMap[beamcodeIndex][3]
            self.beamCodeDict[i] = [beamcodeValue, x, y, z]
            
        just4record0 = self.radacHeaderObj.beamCodeByPulse[0,:]
        
        for i in range(len(list(self.beamCodeDict.values()))):
            xx = numpy.where(just4record0==list(self.beamCodeDict.values())[i][0])
            indexPulseByBeam = self.linear_pulseCount[xx[0]]
            self.beamRangeDict[i] = indexPulseByBeam
    
    def __getExpParameters(self):
        if not(self.status):
            return None
        
        experimentCfgPath = os.path.join(self.path, self.dirnameList[0], 'Setup')
        
        expFinder = glob.glob1(experimentCfgPath,'*.exp')
        if len(expFinder)== 0:
            self.status = 0
            return None
        
        experimentFilename = os.path.join(experimentCfgPath,expFinder[0])
        
        f = open(experimentFilename)
        lines = f.readlines()
        f.close()
        
        parmsList = ['npulsesint*','recordsperfile*','nbeamcodes*','ngates*']
        filterList = [fnmatch.filter(lines, x) for x in parmsList]

        
        values = [re.sub(r'\D',"",x[0]) for x in filterList]
        
        self.npulsesint_fromfile = int(values[0])
        self.recordsperfile_fromfile = int(values[1])
        self.nbeamcodes_fromfile = int(values[2])
        self.ngates_fromfile = int(values[3])
        
        tufileFinder = fnmatch.filter(lines, 'tufile=*')
        tufile = tufileFinder[0].split('=')[1].split('\n')[0]
        tufile = tufile.split('\r')[0]
        tufilename = os.path.join(experimentCfgPath,tufile)
        
        f = open(tufilename)
        lines = f.readlines()
        f.close()
        self.ippSeconds_fromfile = float(lines[1].split()[2])/1E6
        
        
        self.status = 1
    
    def __setIdsAndArrays(self):
        self.dataByFrame = self.__setDataByFrame()
        self.beamCodeByFrame = self.amisrFilePointer.get('Raw11/Data/RadacHeader/BeamCode').value[0, :]        
        self.readRanges()
        self.index_amisr_sample, self.index_amisr_buffer = self.radacHeaderObj.getIndexRangeToPulse(0)
        self.radacTimeByFrame = numpy.zeros(self.radacHeaderObj.npulses)
        if len(self.index_amisr_buffer) > 0:
            self.buffer_radactime = numpy.zeros_like(self.radacTimeByFrame)
        
    
    def __setNextFile(self,online=False):
        
        if not(online):
            newFile = self.__setNextFileOffline()
        else:
            newFile = self.__setNextFileOnline() 
        
        if not(newFile):
            return 0
        
        self.__readHeader()
        
        if self.__firstFile:
            self.__setIdsAndArrays()
            self.__firstFile = False
        
        self.__getBeamCode()
        self.readDataBlock()

    
    def setup(self,path=None,
                    startDate=None, 
                    endDate=None, 
                    startTime=datetime.time(0,0,0), 
                    endTime=datetime.time(23,59,59),
                    walk=True,
                    timezone='ut',
                    all=0,
                    online=False):
        
        self.timezone = timezone
        self.all = all
        self.online = online
        if not(online):
            #Busqueda de archivos offline
            self.searchFilesOffLine(path, startDate, endDate, startTime, endTime, walk)
        else:
            self.searchFilesOnLine(path, walk)
        
        if not(self.filenameList):
            print("There is no files into the folder: %s"%(path))
                
            sys.exit(-1)

        self.__getExpParameters()

        self.fileIndex = -1
        
        self.__setNextFile(online)
        
#         first_beamcode = self.radacHeaderObj.beamCodeByPulse[0,0]
#         index = numpy.where(self.radacHeaderObj.beamCodeByPulse[0,:]!=first_beamcode)[0][0]
        self.profileIndex_offset = self.radacHeaderObj.pulseCount[0,:][0]
        self.profileIndex = self.profileIndex_offset
    
    def readRanges(self):
        dataset = self.amisrFilePointer.get('Raw11/Data/Samples/Range')
        
        self.rangeFromFile = numpy.reshape(dataset.value,(-1))
        return self.rangeFromFile
    
    
    def readRadacTime(self,idrecord, range1, range2):
        self.radacTimeFromFile = self.radacHeaderObj.radacTime.value
        
        radacTimeByFrame = numpy.zeros((self.radacHeaderObj.npulses))
        #radacTimeByFrame = dataset[idrecord - 1,range1]
        #radacTimeByFrame = dataset[idrecord,range2]
        
        return radacTimeByFrame
        
    def readBeamCode(self, idrecord, range1, range2):
        dataset = self.amisrFilePointer.get('Raw11/Data/RadacHeader/BeamCode')
        beamcodeByFrame = numpy.zeros((self.radacHeaderObj.npulses))
        self.beamCodesFromFile = dataset.value
        
        #beamcodeByFrame[range1] = dataset[idrecord - 1, range1]
        #beamcodeByFrame[range2] = dataset[idrecord, range2]
        beamcodeByFrame[range1] = dataset[idrecord, range1]
        beamcodeByFrame[range2] = dataset[idrecord, range2]
        
        return beamcodeByFrame
    
    
    def __setDataByFrame(self):
        ndata = 2 # porque es complejo
        dataByFrame = numpy.zeros((self.radacHeaderObj.npulses, self.radacHeaderObj.nsamples, ndata))
        return dataByFrame
    
    def __readDataSet(self):
        dataset = self.amisrFilePointer.get('Raw11/Data/Samples/Data')
        return dataset

    def __setDataBlock(self,):
        real = self.dataByFrame[:,:,0] #asumo que 0 es real
        imag = self.dataByFrame[:,:,1] #asumo que 1 es imaginario
        datablock = real + imag*1j #armo el complejo
        return datablock
    
    def readSamples_version1(self,idrecord):
        #estas tres primeras lineas solo se deben ejecutar una vez
        if self.flagIsNewFile:
            #reading dataset
            self.dataset = self.__readDataSet()    
            self.flagIsNewFile = 0
        
        if idrecord == 0:
            self.dataByFrame[self.index4_schain_datablock, : ,:] = self.dataset[0, self.index_amisr_sample,:,:]
            self.radacTimeByFrame[self.index4_schain_datablock] = self.radacHeaderObj.radacTime[0, self.index_amisr_sample]
            datablock = self.__setDataBlock()
            if len(self.index_amisr_buffer) > 0:
                self.buffer = self.dataset[0, self.index_amisr_buffer,:,:]
                self.buffer_radactime = self.radacHeaderObj.radacTime[0, self.index_amisr_buffer]
            
            return datablock
        if len(self.index_amisr_buffer) > 0:
            self.dataByFrame[self.index4_buffer,:,:] = self.buffer.copy()
            self.radacTimeByFrame[self.index4_buffer] = self.buffer_radactime.copy()
        self.dataByFrame[self.index4_schain_datablock,:,:] = self.dataset[idrecord, self.index_amisr_sample,:,:]
        self.radacTimeByFrame[self.index4_schain_datablock] = self.radacHeaderObj.radacTime[idrecord, self.index_amisr_sample]
        datablock = self.__setDataBlock()
        if len(self.index_amisr_buffer) > 0:
            self.buffer = self.dataset[idrecord, self.index_amisr_buffer, :, :]
            self.buffer_radactime = self.radacHeaderObj.radacTime[idrecord, self.index_amisr_buffer]    
        
        return datablock
        
    
    def readSamples(self,idrecord):
        if self.flagIsNewFile:
            self.dataByFrame = self.__setDataByFrame()        
            self.beamCodeByFrame = self.amisrFilePointer.get('Raw11/Data/RadacHeader/BeamCode').value[idrecord, :]
                
            #reading ranges
            self.readRanges()
            #reading dataset
            self.dataset = self.__readDataSet()
        
        self.flagIsNewFile = 0
        self.radacTimeByFrame = self.radacHeaderObj.radacTime.value[idrecord, :]
        self.dataByFrame = self.dataset[idrecord, :, :, :]
        datablock = self.__setDataBlock()
        return datablock
    
    
    def readDataBlock(self):
        
        self.datablock = self.readSamples_version1(self.idrecord_count)
        #self.datablock = self.readSamples(self.idrecord_count)
        #print 'record:', self.idrecord_count
        
        self.idrecord_count += 1 
        self.profileIndex = 0
        
        if self.idrecord_count >= self.radacHeaderObj.nrecords:
            self.idrecord_count = 0
            self.flagIsNewFile = 1
    
    def readNextBlock(self):
        
        self.readDataBlock()
        
        if self.flagIsNewFile:
            self.__setNextFile(self.online)
        pass
    
    def __hasNotDataInBuffer(self):
        #self.radacHeaderObj.npulses debe ser otra variable para considerar el numero de pulsos a tomar en el primer y ultimo record
        if self.profileIndex >= self.radacHeaderObj.npulses:
            return 1
        return 0
    
    def printUTC(self):
        print(self.dataOut.utctime)
        print('')
    
    def setObjProperties(self):
        
        self.dataOut.heightList = self.rangeFromFile/1000.0 #km
        self.dataOut.nProfiles = self.radacHeaderObj.npulses
        self.dataOut.nRecords = self.radacHeaderObj.nrecords
        self.dataOut.nBeams = self.radacHeaderObj.nbeams
        self.dataOut.ippSeconds = self.ippSeconds_fromfile
#         self.dataOut.timeInterval = self.dataOut.ippSeconds * self.dataOut.nCohInt
        self.dataOut.frequency = self.frequency_h5file
        self.dataOut.npulseByFrame = self.npulseByFrame
        self.dataOut.nBaud = None
        self.dataOut.nCode = None
        self.dataOut.code = None
        
        self.dataOut.beamCodeDict = self.beamCodeDict
        self.dataOut.beamRangeDict = self.beamRangeDict
        
        if self.timezone == 'lt':
            self.dataOut.timeZone = time.timezone / 60. #get the timezone in minutes
        else: 
            self.dataOut.timeZone = 0 #by default time is UTC
            
    def getData(self):
        
        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            return 0
        
        if self.__hasNotDataInBuffer():
            self.readNextBlock()

        
        if self.datablock is None: # setear esta condicion cuando no hayan datos por leers
            self.dataOut.flagNoData = True 
            return 0
        
        self.dataOut.data = numpy.reshape(self.datablock[self.profileIndex,:],(1,-1))
        
        self.dataOut.utctime = self.radacTimeByFrame[self.profileIndex]
        self.dataOut.profileIndex = self.profileIndex
        self.dataOut.flagNoData = False
        
        self.profileIndex += 1
        
        return self.dataOut.data


    def run(self, **kwargs):
        if not(self.isConfig):
            self.setup(**kwargs)
            self.setObjProperties()
            self.isConfig = True
        
        self.getData()


import numpy

from .jroproc_base import ProcessingUnit, Operation
from schainpy.model.data.jrodata import Correlation

class CorrelationProc(ProcessingUnit):

    pairsList = None

    data_cf = None

    def __init__(self, **kwargs):

        ProcessingUnit.__init__(self, **kwargs)

        self.objectDict = {}
        self.buffer = None
        self.firstdatatime = None
        self.profIndex = 0
        self.dataOut = Correlation()

    def __updateObjFromVoltage(self):

        self.dataOut.timeZone = self.dataIn.timeZone
        self.dataOut.dstFlag = self.dataIn.dstFlag
        self.dataOut.errorCount = self.dataIn.errorCount
        self.dataOut.useLocalTime = self.dataIn.useLocalTime

        self.dataOut.radarControllerHeaderObj = self.dataIn.radarControllerHeaderObj.copy()
        self.dataOut.systemHeaderObj = self.dataIn.systemHeaderObj.copy()
        self.dataOut.channelList = self.dataIn.channelList
        self.dataOut.heightList = self.dataIn.heightList
        self.dataOut.dtype = numpy.dtype([('real','<f4'),('imag','<f4')])
#         self.dataOut.nHeights = self.dataIn.nHeights
#         self.dataOut.nChannels = self.dataIn.nChannels
        self.dataOut.nBaud = self.dataIn.nBaud
        self.dataOut.nCode = self.dataIn.nCode
        self.dataOut.code = self.dataIn.code
#        self.dataOut.nProfiles = self.dataOut.nFFTPoints
        self.dataOut.flagDiscontinuousBlock = self.dataIn.flagDiscontinuousBlock
        self.dataOut.utctime = self.firstdatatime
        self.dataOut.flagDecodeData = self.dataIn.flagDecodeData #asumo q la data esta decodificada
        self.dataOut.flagDeflipData = self.dataIn.flagDeflipData #asumo q la data esta sin flip
        self.dataOut.nCohInt = self.dataIn.nCohInt
#        self.dataOut.nIncohInt = 1
        self.dataOut.ippSeconds = self.dataIn.ippSeconds
        self.dataOut.nProfiles = self.dataIn.nProfiles
        self.dataOut.utctime = self.dataIn.utctime
#        self.dataOut.windowOfFilter = self.dataIn.windowOfFilter

#         self.dataOut.timeInterval = self.dataIn.timeInterval*self.dataOut.nPoints


    def removeDC(self, jspectra):

        nChannel = jspectra.shape[0]

        for i in range(nChannel):
            jspectra_tmp = jspectra[i,:,:]
            jspectra_DC = numpy.mean(jspectra_tmp,axis = 0)

            jspectra_tmp = jspectra_tmp - jspectra_DC
            jspectra[i,:,:] = jspectra_tmp

        return jspectra


    def removeNoise(self, mode = 2):
        indR = numpy.where(self.dataOut.lagR == 0)[0][0]
        indT = numpy.where(self.dataOut.lagT == 0)[0][0]

        jspectra = self.dataOut.data_corr[:,:,indR,:]

        num_chan = jspectra.shape[0]
        num_hei = jspectra.shape[2]

        freq_dc = indT
        ind_vel = numpy.array([-2,-1,1,2]) + freq_dc

        NPot = self.dataOut.getNoise(mode)
        jspectra[:,freq_dc,:] = jspectra[:,freq_dc,:] - NPot
        SPot = jspectra[:,freq_dc,:]
        pairsAutoCorr = self.dataOut.getPairsAutoCorr()
#         self.dataOut.signalPotency = SPot
        self.dataOut.noise = NPot
        self.dataOut.SNR = (SPot/NPot)[pairsAutoCorr]
        self.dataOut.data_corr[:,:,indR,:] = jspectra

        return 1

    def run(self, lags=None, mode = 'time',  pairsList=None, fullBuffer=False, nAvg = 1, removeDC = False, splitCF=False):

        self.dataOut.flagNoData = True

        if self.dataIn.type == "Correlation":

            self.dataOut.copy(self.dataIn)

            return

        if self.dataIn.type == "Voltage":

            nChannels = self.dataIn.nChannels
            nProfiles = self.dataIn.nProfiles
            nHeights = self.dataIn.nHeights
            data_pre = self.dataIn.data

            #---------------    Remover DC    ------------
            if removeDC:
                data_pre = self.removeDC(data_pre)

            #---------------------------------------------
#             pairsList = list(ccfList)
#             for i in acfList:
#                 pairsList.append((i,i))
#
#             ccf_pairs = numpy.arange(len(ccfList))
#             acf_pairs = numpy.arange(len(ccfList),len(pairsList))
            self.__updateObjFromVoltage()
            #----------------------------------------------------------------------
            #Creating temporal buffers
            if fullBuffer:
                tmp = numpy.zeros((len(pairsList), len(lags), nProfiles, nHeights), dtype = 'complex')*numpy.nan
            elif mode == 'time':
                if lags == None:
                    lags = numpy.arange(-nProfiles+1, nProfiles)
                tmp = numpy.zeros((len(pairsList), len(lags), nHeights),dtype='complex')
            elif mode == 'height':
                if lags == None:
                    lags = numpy.arange(-nHeights+1, nHeights)
                tmp = numpy.zeros(len(pairsList), (len(lags), nProfiles),dtype='complex')

            #For loop
            for l in range(len(pairsList)):

                ch0 = pairsList[l][0]
                ch1 = pairsList[l][1]

                for i in range(len(lags)):
                    idx = lags[i]

                    if idx >= 0:
                        if mode == 'time':
                            ccf0 = data_pre[ch0,:nProfiles-idx,:]*numpy.conj(data_pre[ch1,idx:,:]) #time
                        else:
                            ccf0 = data_pre[ch0,:,nHeights-idx]*numpy.conj(data_pre[ch1,:,idx:])   #heights
                    else:
                        if mode == 'time':
                            ccf0 = data_pre[ch0,-idx:,:]*numpy.conj(data_pre[ch1,:nProfiles+idx,:])  #time
                        else:
                            ccf0 = data_pre[ch0,:,-idx:]*numpy.conj(data_pre[ch1,:,:nHeights+idx])   #heights

                    if fullBuffer:
                        tmp[l,i,:ccf0.shape[0],:] = ccf0
                    else:
                        tmp[l,i,:] = numpy.sum(ccf0, axis=0)

            #-----------------------------------------------------------------
            if fullBuffer:
                tmp = numpy.sum(numpy.reshape(tmp,(tmp.shape[0],tmp.shape[1],tmp.shape[2]/nAvg,nAvg,tmp.shape[3])),axis=3)
                self.dataOut.nAvg = nAvg

            self.dataOut.data_cf = tmp
            self.dataOut.mode = mode
            self.dataOut.nLags = len(lags)
            self.dataOut.pairsList = pairsList
            self.dataOut.nPairs = len(pairsList)

            #Se Calcula los factores de Normalizacion
            if mode == 'time':
                delta = self.dataIn.ippSeconds*self.dataIn.nCohInt
            else:
                delta = self.dataIn.heightList[1] - self.dataIn.heightList[0]
            self.dataOut.lagRange = numpy.array(lags)*delta
#             self.dataOut.nCohInt = self.dataIn.nCohInt*nAvg
            self.dataOut.flagNoData = False
#             a = self.dataOut.normFactor
            return



'''
Created on Set 10, 2017

@author: Juan C. Espinoza
'''


import os
import sys
import time
import glob
import datetime

import numpy

from schainpy.model.proc.jroproc_base import ProcessingUnit
from schainpy.model.data.jrodata import Parameters
from schainpy.model.io.jroIO_base import JRODataReader, isNumber
from schainpy.utils import log

FILE_HEADER_STRUCTURE = numpy.dtype([
    ('year', 'f'),
    ('doy', 'f'),
    ('nint', 'f'),
    ('navg', 'f'),
    ('fh', 'f'),
    ('dh', 'f'),
    ('nheights', 'f'),    
    ('ipp', 'f')
])

REC_HEADER_STRUCTURE = numpy.dtype([
    ('magic', 'f'),
    ('hours', 'f'),
    ('interval', 'f'),
    ('h0', 'f'),
    ('nheights', 'f'),
    ('snr1', 'f'),
    ('snr2', 'f'),
    ('snr', 'f'),
])

DATA_STRUCTURE = numpy.dtype([
    ('range', '<u4'),
    ('status', '<u4'),
    ('zonal', '<f4'),
    ('meridional', '<f4'),
    ('vertical', '<f4'),
    ('zonal_a', '<f4'),
    ('meridional_a', '<f4'),
    ('corrected_fading', '<f4'),  # seconds
    ('uncorrected_fading', '<f4'),  # seconds
    ('time_diff', '<f4'),
    ('major_axis', '<f4'),
    ('axial_ratio', '<f4'),
    ('orientation', '<f4'),
    ('sea_power', '<u4'),
    ('sea_algorithm', '<u4')
])


class JULIAParamReader(JRODataReader, ProcessingUnit):
    '''
    Julia data (eej, spf, 150km) *.dat files
    '''

    ext = '.dat'

    def __init__(self, **kwargs):

        ProcessingUnit.__init__(self, **kwargs)

        self.dataOut = Parameters()
        self.counter_records = 0
        self.flagNoMoreFiles = 0
        self.isConfig = False
        self.filename = None
        self.clockpulse = 0.15
        self.kd = 213.6

    def setup(self,
              path=None,
              startDate=None,
              endDate=None,
              ext=None,              
              startTime=datetime.time(0, 0, 0),
              endTime=datetime.time(23, 59, 59),
              timezone=0,
              format=None,
              **kwargs):

        self.path = path
        self.startDate = startDate
        self.endDate = endDate
        self.startTime = startTime
        self.endTime = endTime
        self.datatime = datetime.datetime(1900, 1, 1)
        self.format = format        

        if self.path is None:
            raise ValueError("The path is not valid")

        if ext is None:
            ext = self.ext

        self.search_files(self.path, startDate, endDate, ext)
        self.timezone = timezone
        self.fileIndex = 0

        if not self.fileList:
            log.warning('There is no files matching these date in the folder: {}'.format(
                path), self.name)

        self.setNextFile()

    def search_files(self, path, startDate, endDate, ext):
        '''
         Searching for BLTR rawdata file in path
         Creating a list of file to proces included in [startDate,endDate]

         Input: 
             path - Path to find BLTR rawdata files
             startDate - Select file from this date
             enDate - Select file until this date
             ext - Extension of the file to read
        '''
        
        log.success('Searching files in {} '.format(path), self.name)        
        fileList0 = glob.glob1(path, '{}*{}'.format(self.format.upper(), ext))
        fileList0.sort()

        self.fileList = []
        self.dateFileList = []

        for thisFile in fileList0:            
            year = thisFile[2:4]
            if not isNumber(year):
                continue

            month = thisFile[4:6]
            if not isNumber(month):
                continue

            day = thisFile[6:8]
            if not isNumber(day):
                continue

            year, month, day = int(year), int(month), int(day)
            dateFile = datetime.date(year+2000, month, day)
            
            if (startDate > dateFile) or (endDate < dateFile):
                continue

            self.fileList.append(thisFile)
            self.dateFileList.append(dateFile)
        
        return

    def setNextFile(self):

        file_id = self.fileIndex

        if file_id == len(self.fileList):
            log.success('No more files in the folder', self.name)
            self.flagNoMoreFiles = 1
            return 0
        
        log.success('Opening {}'.format(self.fileList[file_id]), self.name)
        filename = os.path.join(self.path, self.fileList[file_id])

        dirname, name = os.path.split(filename)        
        self.siteFile = name.split('.')[0]
        if self.filename is not None:
            self.fp.close()
        self.filename = filename
        self.fp = open(self.filename, 'rb')
        
        self.header_file = numpy.fromfile(self.fp, FILE_HEADER_STRUCTURE, 1)
        yy = self.header_file['year'] - 1900 * (self.header_file['year'] > 3000)
        self.year = int(yy + 1900 * (yy < 1000))
        self.doy = int(self.header_file['doy'])
        self.dH = round(self.header_file['dh'], 2)
        self.ipp = round(self.header_file['ipp'], 2)
        self.sizeOfFile = os.path.getsize(self.filename)
        self.counter_records = 0
        self.flagIsNewFile = 0
        self.fileIndex += 1

        return 1

    def readNextBlock(self):

        while True:
            if not self.readBlock():
                self.flagIsNewFile = 1
                if not self.setNextFile():
                    return 0            

            if (self.datatime < datetime.datetime.combine(self.startDate, self.startTime)) or \
               (self.datatime > datetime.datetime.combine(self.endDate, self.endTime)):
                log.warning(
                    'Reading Record No. {} -> {} [Skipping]'.format(
                        self.counter_records,
                        self.datatime.ctime()),
                    self.name)
                continue
            break
       
        log.log('Reading Record No. {} -> {}'.format(
            self.counter_records,
            self.datatime.ctime()), self.name)

        return 1

    def readBlock(self):
        
        pointer = self.fp.tell()
        heights, dt = self.readHeader()
        self.fp.seek(pointer)
        buffer_h = []
        buffer_d = []
        while True:
            pointer = self.fp.tell()
            if pointer == self.sizeOfFile:
                return 0
            heights, datatime = self.readHeader()            
            if dt == datatime:                
                buffer_h.append(heights)
                buffer_d.append(self.readData(len(heights)))
                continue
            self.fp.seek(pointer)
            break
        
        if dt.date() > self.datatime.date():
            self.flagDiscontinuousBlock = 1
        self.datatime = dt
        self.time = (dt - datetime.datetime(1970, 1, 1)).total_seconds() + time.timezone        
        self.heights = numpy.concatenate(buffer_h)
        self.buffer = numpy.zeros((5, len(self.heights))) + numpy.nan
        self.buffer[0, :] = numpy.concatenate([buf[0] for buf in buffer_d])
        self.buffer[1, :] = numpy.concatenate([buf[1] for buf in buffer_d])
        self.buffer[2, :] = numpy.concatenate([buf[2] for buf in buffer_d])
        self.buffer[3, :] = numpy.concatenate([buf[3] for buf in buffer_d])
        self.buffer[4, :] = numpy.concatenate([buf[4] for buf in buffer_d])
        
        self.counter_records += 1

        return 1

    def readHeader(self):
        '''
        Parse recordHeader
        '''
        
        self.header_rec = numpy.fromfile(self.fp, REC_HEADER_STRUCTURE, 1)
        self.interval = self.header_rec['interval']
        if self.header_rec['magic'] == 888.:
            self.header_rec['h0'] = round(self.header_rec['h0'], 2)
            nheights = int(self.header_rec['nheights'])            
            hours = float(self.header_rec['hours'][0])
            heights = numpy.arange(nheights) * self.dH + self.header_rec['h0']
            datatime = datetime.datetime(self.year, 1, 1) + datetime.timedelta(days=self.doy-1, hours=hours)            
            return heights, datatime
        else:
            return False
        
    def readData(self, N):
        '''
        Parse data
        '''

        buffer = numpy.fromfile(self.fp, 'f', 8*N).reshape(N, 8)
        
        pow0 = buffer[:, 0]
        pow1 = buffer[:, 1]
        acf0 = (buffer[:,2] + buffer[:,3]*1j) / pow0
        acf1 = (buffer[:,4] + buffer[:,5]*1j) / pow1
        dccf = (buffer[:,6] + buffer[:,7]*1j) / (pow0*pow1)

        ### SNR
        sno = (pow0 + pow1 - self.header_rec['snr']) / self.header_rec['snr']
        sno10 = numpy.log10(sno)
        # dsno = 1.0 / numpy.sqrt(self.header_file['nint'] * self.header_file['navg']) * (1 + (1 / sno))
        
        ### Vertical Drift
        sp = numpy.sqrt(numpy.abs(acf0)*numpy.abs(acf1))
        sp[numpy.where(numpy.abs(sp) >= 1.0)] = numpy.sqrt(0.9999)
                    
        vzo = -numpy.arctan2(acf0.imag + acf1.imag,acf0.real + acf1.real)*1.5E5*1.5/(self.ipp*numpy.pi)
        dvzo = numpy.sqrt(1.0 - sp*sp)*0.338*1.5E5/(numpy.sqrt(self.header_file['nint']*self.header_file['navg'])*sp*self.ipp)            
        err = numpy.where(dvzo <= 0.1)
        dvzo[err] = 0.1

        #Zonal Drifts
        dt = self.header_file['nint']*self.ipp / 1.5E5
        coh = numpy.sqrt(numpy.abs(dccf))
        err = numpy.where(coh >= 1.0)            
        coh[err] = numpy.sqrt(0.99999)
            
        err = numpy.where(coh <= 0.1)
        coh[err] = numpy.sqrt(0.1)
                
        vxo = numpy.arctan2(dccf.imag, dccf.real)*self.header_rec['h0']*1.0E3/(self.kd*dt)
        dvxo = numpy.sqrt(1.0 - coh*coh)*self.header_rec['h0']*1.0E3/(numpy.sqrt(self.header_file['nint']*self.header_file['navg'])*coh*self.kd*dt)
        
        err = numpy.where(dvxo <= 0.1)            
        dvxo[err] = 0.1

        return vzo, dvzo, vxo, dvxo, sno10 

    def set_output(self):
        '''
        Storing data from databuffer to dataOut object
        '''
        
        self.dataOut.data_SNR = self.buffer[4].reshape(1, -1)
        self.dataOut.heightList = self.heights
        self.dataOut.data_param = self.buffer[0:4,]
        self.dataOut.utctimeInit = self.time
        self.dataOut.utctime = self.time
        self.dataOut.useLocalTime = True
        self.dataOut.paramInterval = self.interval
        self.dataOut.timezone = self.timezone
        self.dataOut.sizeOfFile = self.sizeOfFile
        self.dataOut.flagNoData = False
        self.dataOut.flagDiscontinuousBlock = self.flagDiscontinuousBlock

    def getData(self):
        '''
        Storing data from databuffer to dataOut object
        '''
        if self.flagNoMoreFiles:
            self.dataOut.flagNoData = True
            log.success('No file left to process', self.name)
            return 0

        if not self.readNextBlock():
            self.dataOut.flagNoData = True
            return 0

        self.set_output()

        return 1


'''
Created on Jul 9, 2014

@author: roj-idl71
'''
import os
import datetime
import numpy

from schainpy.model.graphics.jroplot_base import Plot, plt


class ScopePlot(Plot):

    '''
       Plot for Scope
    '''

    CODE = 'scope'
    plot_type = 'scatter'

    def setup(self):

        self.xaxis = 'Range (Km)'
        self.ncols = 1
        self.nrows = 1
        self.nplots = 1
        self.ylabel = 'Intensity [dB]'
        self.titles = ['Scope']
        self.colorbar = False
        self.width = 6
        self.height = 4

    def plot_iq(self, x, y, channelIndexList, thisDatetime, wintitle):

        yreal = y[channelIndexList,:].real
        yimag = y[channelIndexList,:].imag
        title = wintitle + " Scope: %s" %(thisDatetime.strftime("%d-%b-%Y"))
        self.xlabel = "Range (Km)"
        self.ylabel = "Intensity - IQ"

        self.y = yreal
        self.x = x
        self.xmin = min(x)
        self.xmax = max(x)


        self.titles[0] = title

        for i,ax in enumerate(self.axes):
            title = "Channel %d" %(i)
            if ax.firsttime:
                ax.plt_r = ax.plot(x, yreal[i,:], color='b')[0]
                ax.plt_i = ax.plot(x, yimag[i,:], color='r')[0]
            else:
                ax.plt_r.set_data(x, yreal[i,:])
                ax.plt_i.set_data(x, yimag[i,:])

    def plot_power(self, x, y, channelIndexList, thisDatetime, wintitle):
        y = y[channelIndexList,:] * numpy.conjugate(y[channelIndexList,:])
        yreal = y.real
        yreal = 10*numpy.log10(yreal)
        self.y = yreal
        title = wintitle + " Scope: %s" %(thisDatetime.strftime("%d-%b-%Y"))
        self.xlabel = "Range (Km)"
        self.ylabel = "Intensity"
        self.xmin = min(x)
        self.xmax = max(x)


        self.titles[0] = title

        for i,ax in enumerate(self.axes):
            title = "Channel %d" %(i)

            ychannel = yreal[i,:]

            if ax.firsttime:
                ax.plt_r = ax.plot(x, ychannel)[0]
            else:
                #pass
                ax.plt_r.set_data(x, ychannel)

    def plot_weatherpower(self, x, y, channelIndexList, thisDatetime, wintitle):


        y      = y[channelIndexList,:]
        yreal  = y.real
        yreal  = 10*numpy.log10(yreal)
        self.y = yreal
        title  = wintitle + " Scope: %s" %(thisDatetime.strftime("%d-%b-%Y %H:%M:%S"))
        self.xlabel = "Range (Km)"
        self.ylabel = "Intensity"
        self.xmin   = min(x)
        self.xmax   = max(x)

        self.titles[0] =title
        for i,ax in enumerate(self.axes):
            title    = "Channel %d" %(i)

            ychannel = yreal[i,:]

            if ax.firsttime:
                ax.plt_r = ax.plot(x, ychannel)[0]
            else:
                #pass
                ax.plt_r.set_data(x, ychannel)

    def plot_weathervelocity(self, x, y, channelIndexList, thisDatetime, wintitle):

        x = x[channelIndexList,:]
        yreal  = y
        self.y = yreal
        title = wintitle + " Scope: %s" %(thisDatetime.strftime("%d-%b-%Y %H:%M:%S"))
        self.xlabel = "Velocity (m/s)"
        self.ylabel = "Range (Km)"
        self.xmin   = numpy.min(x)
        self.xmax   = numpy.max(x)
        self.titles[0] =title
        for i,ax in enumerate(self.axes):
            title    = "Channel %d" %(i)
            xchannel    = x[i,:]
            if ax.firsttime:
                ax.plt_r = ax.plot(xchannel, yreal)[0]
            else:
                #pass
                ax.plt_r.set_data(xchannel, yreal)

    def plot_weatherspecwidth(self, x, y, channelIndexList, thisDatetime, wintitle):

        x = x[channelIndexList,:]
        yreal  = y
        self.y = yreal
        title = wintitle + " Scope: %s" %(thisDatetime.strftime("%d-%b-%Y %H:%M:%S"))
        self.xlabel = "width "
        self.ylabel = "Range (Km)"
        self.xmin   = numpy.min(x)
        self.xmax   = numpy.max(x)
        self.titles[0] =title
        for i,ax in enumerate(self.axes):
            title    = "Channel %d" %(i)
            xchannel    = x[i,:]
            if ax.firsttime:
                ax.plt_r = ax.plot(xchannel, yreal)[0]
            else:
                #pass
                ax.plt_r.set_data(xchannel, yreal)

    def plot(self):
        if self.channels:
            channels = self.channels
        else:
            channels = self.data.channels

        thisDatetime = datetime.datetime.utcfromtimestamp(self.data.times[-1])
        if self.CODE == "pp_power":
            scope = self.data['pp_power']
        elif self.CODE == "pp_signal":
            scope = self.data["pp_signal"]
        elif self.CODE == "pp_velocity":
            scope = self.data["pp_velocity"]
        elif self.CODE == "pp_specwidth":
            scope = self.data["pp_specwidth"]
        else:
            scope =self.data["scope"]

        if self.data.flagDataAsBlock:

            for i in range(self.data.nProfiles):

                wintitle1 = " [Profile = %d] " %i
                if self.CODE =="scope":
                    if self.type == "power":
                        self.plot_power(self.data.heights,
                                        scope[:,i,:],
                                        channels,
                                        thisDatetime,
                                        wintitle1
                                        )

                    if self.type == "iq":
                        self.plot_iq(self.data.heights,
                                     scope[:,i,:],
                                     channels,
                                     thisDatetime,
                                     wintitle1
                                    )
                if self.CODE=="pp_power":
                    self.plot_weatherpower(self.data.heights,
                               scope[:,i,:],
                               channels,
                               thisDatetime,
                               wintitle
                               )
                if self.CODE=="pp_signal":
                    self.plot_weatherpower(self.data.heights,
                               scope[:,i,:],
                               channels,
                               thisDatetime,
                               wintitle
                               )
                if self.CODE=="pp_velocity":
                    self.plot_weathervelocity(scope[:,i,:],
                               self.data.heights,
                               channels,
                               thisDatetime,
                               wintitle
                               )
                if self.CODE=="pp_spcwidth":
                    self.plot_weatherspecwidth(scope[:,i,:],
                               self.data.heights,
                               channels,
                               thisDatetime,
                               wintitle
                               )
        else:
            wintitle = " [Profile = %d] " %self.data.profileIndex
            if self.CODE== "scope":
                if self.type == "power":
                     self.plot_power(self.data.heights,
                                scope,
                                channels,
                                thisDatetime,
                                wintitle
                                )

                if self.type == "iq":
                    self.plot_iq(self.data.heights,
                                scope,
                                channels,
                                thisDatetime,
                                wintitle
                                )
            if self.CODE=="pp_power":
                self.plot_weatherpower(self.data.heights,
                                    scope,
                                    channels,
                                    thisDatetime,
                                    wintitle
                                       )
            if self.CODE=="pp_signal":
                self.plot_weatherpower(self.data.heights,
                                    scope,
                                    channels,
                                    thisDatetime,
                                    wintitle
                                       )
            if self.CODE=="pp_velocity":
                self.plot_weathervelocity(scope,
                                       self.data.heights,
                                       channels,
                                       thisDatetime,
                                       wintitle
                                       )
            if self.CODE=="pp_specwidth":
                self.plot_weatherspecwidth(scope,
                                       self.data.heights,
                                       channels,
                                       thisDatetime,
                                       wintitle
                                       )



class PulsepairPowerPlot(ScopePlot):
    '''
    Plot for  P= S+N
    '''

    CODE = 'pp_power'
    plot_type = 'scatter'
    buffering = False

class PulsepairVelocityPlot(ScopePlot):
    '''
    Plot for VELOCITY
    '''
    CODE = 'pp_velocity'
    plot_type = 'scatter'
    buffering = False

class PulsepairSpecwidthPlot(ScopePlot):
    '''
    Plot for WIDTH
    '''
    CODE = 'pp_specwidth'
    plot_type = 'scatter'
    buffering = False

class PulsepairSignalPlot(ScopePlot):
    '''
    Plot for S
    '''

    CODE = 'pp_signal'
    plot_type = 'scatter'
    buffering = False



import os
import datetime
import numpy
import copy
from schainpy.model.graphics.jroplot_base import Plot


class CorrelationPlot(Plot):
    isConfig = None
    __nsubplots = None

    WIDTHPROF = None
    HEIGHTPROF = None
    PREFIX = 'corr'

    def __init__(self, **kwargs):
        Figure.__init__(self, **kwargs)
        self.isConfig = False
        self.__nsubplots = 1

        self.WIDTH = 280
        self.HEIGHT = 250
        self.WIDTHPROF = 120
        self.HEIGHTPROF = 0
        self.counter_imagwr = 0

        self.PLOT_CODE = 1
        self.FTP_WEI = None
        self.EXP_CODE = None
        self.SUB_EXP_CODE = None
        self.PLOT_POS = None

    def getSubplots(self):

        ncol = int(numpy.sqrt(self.nplots)+0.9)
        nrow = int(self.nplots*1./ncol + 0.9)

        return nrow, ncol

    def setup(self, id, nplots, wintitle, showprofile=False, show=True):

        showprofile = False
        self.__showprofile = showprofile
        self.nplots = nplots

        ncolspan = 1
        colspan = 1
        if showprofile:
            ncolspan = 3
            colspan = 2
            self.__nsubplots = 2

        self.createFigure(id = id,
                          wintitle = wintitle,
                          widthplot = self.WIDTH + self.WIDTHPROF,
                          heightplot = self.HEIGHT + self.HEIGHTPROF,
                          show=show)

        nrow, ncol = self.getSubplots()

        counter = 0
        for y in range(nrow):
            for x in range(ncol):

                if counter >= self.nplots:
                    break

                self.addAxes(nrow, ncol*ncolspan, y, x*ncolspan, colspan, 1)

                if showprofile:
                    self.addAxes(nrow, ncol*ncolspan, y, x*ncolspan+colspan, 1, 1)

                counter += 1

    def run(self, dataOut, id, wintitle="", channelList=None, showprofile=False,
            xmin=None, xmax=None, ymin=None, ymax=None, zmin=None, zmax=None,
            save=False, figpath='./', figfile=None, show=True, ftp=False, wr_period=1,
            server=None, folder=None, username=None, password=None,
            ftp_wei=0, exp_code=0, sub_exp_code=0, plot_pos=0, realtime=False):

        """

        Input:
            dataOut         :
            id        :
            wintitle        :
            channelList     :
            showProfile     :
            xmin            :    None,
            xmax            :    None,
            ymin            :    None,
            ymax            :    None,
            zmin            :    None,
            zmax            :    None
        """

        if dataOut.flagNoData:
            return None

        if realtime:
            if not(isRealtime(utcdatatime = dataOut.utctime)):
                print('Skipping this plot function')
                return

        if channelList == None:
            channelIndexList = dataOut.channelIndexList
        else:
            channelIndexList = []
            for channel in channelList:
                if channel not in dataOut.channelList:
                    raise ValueError("Channel %d is not in dataOut.channelList")
                channelIndexList.append(dataOut.channelList.index(channel))

        factor = dataOut.normFactor
        lenfactor = factor.shape[1]
        x = dataOut.getLagTRange(1)
        y = dataOut.getHeiRange()

        z = copy.copy(dataOut.data_corr[:,:,0,:])
        for i in range(dataOut.data_corr.shape[0]):
            z[i,:,:] = z[i,:,:]/factor[i,:]
        zdB = numpy.abs(z)

        avg = numpy.average(z, axis=1)
#         avg = numpy.nanmean(z, axis=1)
#         noise = dataOut.noise/factor

        #thisDatetime = dataOut.datatime
        thisDatetime = datetime.datetime.utcfromtimestamp(dataOut.getTimeRange()[0])
        title = wintitle + " Correlation"
        xlabel = "Lag T (s)"
        ylabel = "Range (Km)"

        if not self.isConfig:

            nplots = dataOut.data_corr.shape[0]

            self.setup(id=id,
                       nplots=nplots,
                       wintitle=wintitle,
                       showprofile=showprofile,
                       show=show)

            if xmin == None: xmin = numpy.nanmin(x)
            if xmax == None: xmax = numpy.nanmax(x)
            if ymin == None: ymin = numpy.nanmin(y)
            if ymax == None: ymax = numpy.nanmax(y)
            if zmin == None: zmin = 0
            if zmax == None: zmax = 1

            self.FTP_WEI = ftp_wei
            self.EXP_CODE = exp_code
            self.SUB_EXP_CODE = sub_exp_code
            self.PLOT_POS = plot_pos

            self.isConfig = True

        self.setWinTitle(title)

        for i in range(self.nplots):
            str_datetime = '%s %s'%(thisDatetime.strftime("%Y/%m/%d"),thisDatetime.strftime("%H:%M:%S"))
            title = "Channel %d and %d: : %s" %(dataOut.pairsList[i][0],dataOut.pairsList[i][1] , str_datetime)
            axes = self.axesList[i*self.__nsubplots]
            axes.pcolor(x, y, zdB[i,:,:],
                        xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, zmin=zmin, zmax=zmax,
                        xlabel=xlabel, ylabel=ylabel, title=title,
                        ticksize=9, cblabel='')

#             if self.__showprofile:
#                 axes = self.axesList[i*self.__nsubplots +1]
#                 axes.pline(avgdB[i], y,
#                         xmin=zmin, xmax=zmax, ymin=ymin, ymax=ymax,
#                         xlabel='dB', ylabel='', title='',
#                         ytick_visible=False,
#                         grid='x')
#
#                 noiseline = numpy.repeat(noisedB[i], len(y))
#                 axes.addpline(noiseline, y, idline=1, color="black", linestyle="dashed", lw=2)

        self.draw()

        self.save(figpath=figpath,
                  figfile=figfile,
                  save=save,
                  ftp=ftp,
                  wr_period=wr_period,
                  thisDatetime=thisDatetime)


import os
import datetime
import numpy

from schainpy.model.graphics.jroplot_base import Plot, plt
from schainpy.model.graphics.jroplot_spectra import SpectraPlot, RTIPlot, CoherencePlot
from schainpy.utils import log

EARTH_RADIUS = 6.3710e3


def ll2xy(lat1, lon1, lat2, lon2):

    p = 0.017453292519943295
    a = 0.5 - numpy.cos((lat2 - lat1) * p)/2 + numpy.cos(lat1 * p) * \
        numpy.cos(lat2 * p) * (1 - numpy.cos((lon2 - lon1) * p)) / 2
    r = 12742 * numpy.arcsin(numpy.sqrt(a))
    theta = numpy.arctan2(numpy.sin((lon2-lon1)*p)*numpy.cos(lat2*p), numpy.cos(lat1*p)
                          * numpy.sin(lat2*p)-numpy.sin(lat1*p)*numpy.cos(lat2*p)*numpy.cos((lon2-lon1)*p))
    theta = -theta + numpy.pi/2
    return r*numpy.cos(theta), r*numpy.sin(theta)


def km2deg(km):
    '''
    Convert distance in km to degrees
    '''

    return numpy.rad2deg(km/EARTH_RADIUS)



class SpectralMomentsPlot(SpectraPlot):
    '''
    Plot for Spectral Moments
    '''
    CODE = 'spc_moments'
    colormap = 'jet'
    plot_type = 'pcolor'


class SnrPlot(RTIPlot):
    '''
    Plot for SNR Data
    '''

    CODE = 'snr'
    colormap = 'jet'


class DopplerPlot(RTIPlot):
    '''
    Plot for DOPPLER Data (1st moment)
    '''

    CODE = 'dop'
    colormap = 'jet'


class PowerPlot(RTIPlot):
    '''
    Plot for Power Data (0 moment)
    '''

    CODE = 'pow'
    colormap = 'jet'


class SpectralWidthPlot(RTIPlot):
    '''
    Plot for Spectral Width Data (2nd moment)
    '''

    CODE = 'width'
    colormap = 'jet'


class SkyMapPlot(Plot):
    '''
    Plot for meteors detection data
    '''

    CODE = 'param'

    def setup(self):

        self.ncols = 1
        self.nrows = 1
        self.width = 7.2
        self.height = 7.2
        self.nplots = 1
        self.xlabel = 'Zonal Zenith Angle (deg)'
        self.ylabel = 'Meridional Zenith Angle (deg)'
        self.polar = True
        self.ymin = -180
        self.ymax = 180
        self.colorbar = False

    def plot(self):

        arrayParameters = numpy.concatenate(self.data['param'])
        error = arrayParameters[:, -1]
        indValid = numpy.where(error == 0)[0]
        finalMeteor = arrayParameters[indValid, :]
        finalAzimuth = finalMeteor[:, 3]
        finalZenith = finalMeteor[:, 4]

        x = finalAzimuth * numpy.pi / 180
        y = finalZenith

        ax = self.axes[0]

        if ax.firsttime:
            ax.plot = ax.plot(x, y, 'bo', markersize=5)[0]
        else:
            ax.plot.set_data(x, y)

        dt1 = self.getDateTime(self.data.min_time).strftime('%y/%m/%d %H:%M:%S')
        dt2 = self.getDateTime(self.data.max_time).strftime('%y/%m/%d %H:%M:%S')
        title = 'Meteor Detection Sky Map\n %s - %s \n Number of events: %5.0f\n' % (dt1,
                                                                                     dt2,
                                                                                     len(x))
        self.titles[0] = title


class ParametersPlot(RTIPlot):
    '''
    Plot for data_param object
    '''

    CODE = 'param'
    colormap = 'seismic'

    def setup(self):
        self.xaxis = 'time'
        self.ncols = 1
        self.nrows = self.data.shape(self.CODE)[0]
        self.nplots = self.nrows
        self.plots_adjust.update({'hspace':0.8, 'left': 0.1, 'bottom': 0.08, 'right':0.95, 'top': 0.95})
        
        if not self.xlabel:
            self.xlabel = 'Time'
        
        if self.showSNR:
            self.nrows += 1
            self.nplots += 1

        self.ylabel = 'Height [km]'
        if not self.titles:
            self.titles = self.data.parameters \
                if self.data.parameters else ['Param {}'.format(x) for x in range(self.nrows)]
            if self.showSNR:
                self.titles.append('SNR')

    def plot(self):
        self.data.normalize_heights()
        self.x = self.data.times
        self.y = self.data.heights
        if self.showSNR:
            self.z = numpy.concatenate(
                (self.data[self.CODE], self.data['snr'])
            )
        else:
            self.z = self.data[self.CODE]

        self.z = numpy.ma.masked_invalid(self.z)

        if self.decimation is None:
            x, y, z = self.fill_gaps(self.x, self.y, self.z)
        else:
            x, y, z = self.fill_gaps(*self.decimate())

        for n, ax in enumerate(self.axes):

            self.zmax = self.zmax if self.zmax is not None else numpy.max(
                self.z[n])
            self.zmin = self.zmin if self.zmin is not None else numpy.min(
                self.z[n])

            if ax.firsttime:
                if self.zlimits is not None:
                    self.zmin, self.zmax = self.zlimits[n]

                ax.plt = ax.pcolormesh(x, y, z[n].T * self.factors[n],
                                       vmin=self.zmin,
                                       vmax=self.zmax,
                                       cmap=self.cmaps[n]
                                       )
            else:
                if self.zlimits is not None:
                    self.zmin, self.zmax = self.zlimits[n]
                ax.collections.remove(ax.collections[0])
                ax.plt = ax.pcolormesh(x, y, z[n].T * self.factors[n],
                                       vmin=self.zmin,
                                       vmax=self.zmax,
                                       cmap=self.cmaps[n]
                                       )


class OutputPlot(ParametersPlot):
    '''
    Plot data_output object
    '''

    CODE = 'output'
    colormap = 'seismic'


class PolarMapPlot(Plot):
    '''
    Plot for weather radar
    '''

    CODE = 'param'
    colormap = 'seismic'

    def setup(self):
        self.ncols = 1
        self.nrows = 1
        self.width = 9
        self.height = 8
        self.mode = self.data.meta['mode']
        if self.channels is not None:
            self.nplots = len(self.channels)
            self.nrows = len(self.channels)
        else:
            self.nplots = self.data.shape(self.CODE)[0]
            self.nrows = self.nplots
            self.channels = list(range(self.nplots))
        if self.mode == 'E':
            self.xlabel = 'Longitude'
            self.ylabel = 'Latitude'
        else:
            self.xlabel = 'Range (km)'
            self.ylabel = 'Height (km)'
        self.bgcolor = 'white'
        self.cb_labels = self.data.meta['units']
        self.lat = self.data.meta['latitude']
        self.lon = self.data.meta['longitude']
        self.xmin, self.xmax = float(
            km2deg(self.xmin) + self.lon), float(km2deg(self.xmax) + self.lon)
        self.ymin, self.ymax = float(
            km2deg(self.ymin) + self.lat), float(km2deg(self.ymax) + self.lat)
        #  self.polar = True

    def plot(self):

        for n, ax in enumerate(self.axes):
            data = self.data['param'][self.channels[n]]

            zeniths = numpy.linspace(
                0, self.data.meta['max_range'], data.shape[1])
            if self.mode == 'E':
                azimuths = -numpy.radians(self.data.heights)+numpy.pi/2
                r, theta = numpy.meshgrid(zeniths, azimuths)
                x, y = r*numpy.cos(theta)*numpy.cos(numpy.radians(self.data.meta['elevation'])), r*numpy.sin(
                    theta)*numpy.cos(numpy.radians(self.data.meta['elevation']))
                x = km2deg(x) + self.lon
                y = km2deg(y) + self.lat
            else:
                azimuths = numpy.radians(self.data.heights)
                r, theta = numpy.meshgrid(zeniths, azimuths)
                x, y = r*numpy.cos(theta), r*numpy.sin(theta)
            self.y = zeniths

            if ax.firsttime:
                if self.zlimits is not None:
                    self.zmin, self.zmax = self.zlimits[n]
                ax.plt = ax.pcolormesh(  # r, theta, numpy.ma.array(data, mask=numpy.isnan(data)),
                    x, y, numpy.ma.array(data, mask=numpy.isnan(data)),
                    vmin=self.zmin,
                    vmax=self.zmax,
                    cmap=self.cmaps[n])
            else:
                if self.zlimits is not None:
                    self.zmin, self.zmax = self.zlimits[n]
                ax.collections.remove(ax.collections[0])
                ax.plt = ax.pcolormesh(  # r, theta, numpy.ma.array(data, mask=numpy.isnan(data)),
                    x, y, numpy.ma.array(data, mask=numpy.isnan(data)),
                    vmin=self.zmin,
                    vmax=self.zmax,
                    cmap=self.cmaps[n])

            if self.mode == 'A':
                continue

            # plot district names
            f = open('/data/workspace/schain_scripts/distrito.csv')
            for line in f:
                label, lon, lat = [s.strip() for s in line.split(',') if s]
                lat = float(lat)
                lon = float(lon)
                # ax.plot(lon, lat, '.b', ms=2)
                ax.text(lon, lat, label.decode('utf8'), ha='center',
                        va='bottom', size='8', color='black')

            # plot limites
            limites = []
            tmp = []
            for line in open('/data/workspace/schain_scripts/lima.csv'):
                if '#' in line:
                    if tmp:
                        limites.append(tmp)
                    tmp = []
                    continue
                values = line.strip().split(',')
                tmp.append((float(values[0]), float(values[1])))
            for points in limites:
                ax.add_patch(
                    Polygon(points, ec='k', fc='none', ls='--', lw=0.5))

            # plot Cuencas
            for cuenca in ('rimac', 'lurin', 'mala', 'chillon', 'chilca', 'chancay-huaral'):
                f = open('/data/workspace/schain_scripts/{}.csv'.format(cuenca))
                values = [line.strip().split(',') for line in f]
                points = [(float(s[0]), float(s[1])) for s in values]
                ax.add_patch(Polygon(points, ec='b', fc='none'))

            # plot grid
            for r in (15, 30, 45, 60):
                ax.add_artist(plt.Circle((self.lon, self.lat),
                                         km2deg(r), color='0.6', fill=False, lw=0.2))
                ax.text(
                    self.lon + (km2deg(r))*numpy.cos(60*numpy.pi/180),
                    self.lat + (km2deg(r))*numpy.sin(60*numpy.pi/180),
                    '{}km'.format(r),
                    ha='center', va='bottom', size='8', color='0.6', weight='heavy')

        if self.mode == 'E':
            title = 'El={}$^\circ$'.format(self.data.meta['elevation'])
            label = 'E{:02d}'.format(int(self.data.meta['elevation']))
        else:
            title = 'Az={}$^\circ$'.format(self.data.meta['azimuth'])
            label = 'A{:02d}'.format(int(self.data.meta['azimuth']))

        self.save_labels = ['{}-{}'.format(lbl, label) for lbl in self.labels]
        self.titles = ['{} {}'.format(
            self.data.parameters[x], title) for x in self.channels]

